from pathlib import Path
from typing import Optional
import numpy as np
from qtpy.QtCore import Qt

from allencell_ml_segmenter._style import Style
from allencell_ml_segmenter.core.dialog_box import DialogBox
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.core.file_input_widget import (
    FileInputWidget,
    FileInputModel,
)
from allencell_ml_segmenter.prediction.model import (
    PredictionModel,
)

from allencell_ml_segmenter.core.file_input_model import InputMode
from allencell_ml_segmenter.prediction.service import ModelFileService
from allencell_ml_segmenter.core.view import View, MainWindow
from allencell_ml_segmenter.prediction.file_write_progress_tracker import (
    FileWriteProgressTracker,
)
from allencell_ml_segmenter.utils.file_utils import FileUtils
from qtpy.QtWidgets import (
    QVBoxLayout,
    QSizePolicy,
    QPushButton,
    QFrame,
    QLabel,
)
from allencell_ml_segmenter.main.i_viewer import IViewer
from allencell_ml_segmenter.core.image_data_extractor import (
    IImageDataExtractor,
    AICSImageDataExtractor,
)


class PredictionView(View, MainWindow):
    """
    Holds the image and model input widgets for prediction.
    """

    def __init__(
        self,
        main_model: MainModel,
        prediction_model: PredictionModel,
        file_input_model: FileInputModel,
        viewer: IViewer,
        img_data_extractor: IImageDataExtractor = AICSImageDataExtractor.global_instance(),
    ):
        super().__init__()
        self._main_model: MainModel = main_model
        self._prediction_model: PredictionModel = prediction_model
        self._file_input_model: FileInputModel = file_input_model
        self._viewer: IViewer = viewer
        self._img_data_extractor = img_data_extractor

        self._service: ModelFileService = ModelFileService(
            self._file_input_model
        )

        layout: QVBoxLayout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum
        )

        self._title: QLabel = QLabel("SEGMENTATION PREDICTION", self)
        self._title.setObjectName("title")
        layout.addWidget(self._title, alignment=Qt.AlignmentFlag.AlignHCenter)

        self._file_input_widget: FileInputWidget = FileInputWidget(
            self._file_input_model, self._viewer, self._service
        )
        self._file_input_widget.setObjectName("fileInput")

        # Disabled for V1 chrishu 3/30/24 https://github.com/AllenCell/allencell-ml-segmenter/issues/274
        # self._model_input_widget: ModelInputWidget = ModelInputWidget(
        #     self._prediction_model
        # )
        # self._model_input_widget.setObjectName("modelInput")

        # Dummy divs allow for easy alignment
        top_container: QVBoxLayout = QVBoxLayout()
        top_dummy: QFrame = QFrame()
        bottom_container: QVBoxLayout = QVBoxLayout()
        bottom_dummy: QFrame = QFrame()

        top_container.addWidget(self._file_input_widget)
        top_dummy.setLayout(top_container)
        layout.addWidget(top_dummy)

        # Disabled for V1 chrishu 3/30/24 https://github.com/AllenCell/allencell-ml-segmenter/issues/274
        # bottom_container.addWidget(self._model_input_widget)

        bottom_dummy.setLayout(bottom_container)
        layout.addWidget(bottom_dummy)

        self._run_btn: QPushButton = QPushButton("Run")
        self._run_btn.setObjectName("run")
        layout.addWidget(self._run_btn)
        self._run_btn.clicked.connect(self.run_btn_handler)

        self.setStyleSheet(Style.get_stylesheet("prediction_view.qss"))

        self._main_model.subscribe(
            Event.PROCESS_TRAINING_COMPLETE,
            self,
            lambda e: self._main_model.set_current_view(self),
        )

    def run_btn_handler(self) -> None:
        # dispatch events to set _prediction_model._input_image_path to a real CSV

        # get image paths from napari if they are selected
        self._prediction_model.dispatch_prediction_get_image_paths_from_napari()
        # Verify prediction is able to start, and write csv if needed
        self._prediction_model.dispatch_prediction_setup()

        total_num_images: Optional[int] = (
            self._prediction_model.get_total_num_images()
        )
        output_seg_dir: Optional[Path] = (
            self._file_input_model.get_output_seg_directory()
        )
        if total_num_images is not None and output_seg_dir is not None:
            progress_tracker: FileWriteProgressTracker = (
                FileWriteProgressTracker(
                    output_seg_dir,
                    total_num_images,
                )
            )
            self.startLongTaskWithProgressBar(progress_tracker)

    def doWork(self) -> None:
        self._prediction_model.dispatch_prediction()
        # TODO Need way to set result images to show after prediction complete and refresh viewer.

    def getTypeOfWork(self) -> str:
        return "Prediction"

    def showResults(self) -> None:
        output_path: Optional[Path] = (
            self._file_input_model.get_output_seg_directory()
        )

        # Display images if prediction inputs are from Napari Layers
        if (
            self._file_input_model.get_input_mode()
            == InputMode.FROM_NAPARI_LAYERS
            and output_path is not None
        ):
            raw_imgs: Optional[list[Path]] = (
                self._file_input_model.get_selected_paths()
            )
            segmentations: list[Path] = (
                FileUtils.get_all_files_in_dir_ignore_hidden(output_path)
            )
            channel: Optional[int] = (
                self._file_input_model.get_image_input_channel_index()
            )
            if raw_imgs is None or channel is None:
                raise RuntimeError("Insufficient data to show results")

            # here, we will pair raw images and segmentations based on the stem component of their paths
            stem_to_data: dict[str, dict[str, Path]] = {
                raw_img.stem: {"raw": raw_img} for raw_img in raw_imgs
            }

            if segmentations:
                for seg in segmentations:
                    # ignore files in the folder that aren't from most recent predictions
                    if seg.stem in stem_to_data:
                        stem_to_data[seg.stem]["seg"] = seg

                self._viewer.clear_layers()
                for data in stem_to_data.values():
                    raw_np_data: Optional[np.ndarray] = (
                        self._img_data_extractor.extract_image_data(
                            data["raw"], channel=channel
                        ).np_data
                    )
                    seg_np_data: Optional[np.ndarray] = (
                        self._img_data_extractor.extract_image_data(
                            data["seg"], seg=1
                        ).np_data
                    )
                    if raw_np_data is not None:
                        self._viewer.add_image(
                            raw_np_data,
                            name=f"[raw] {data['raw'].name}",
                            metadata={"source_path": str(data["raw"])},
                        )
                    if seg_np_data is not None:
                        self._viewer.add_labels(
                            seg_np_data,
                            name=f"[seg] {data['seg'].name}",
                            metadata={
                                "source_path": data["seg"],
                                "prob_map": self._img_data_extractor.extract_image_data(
                                    data["seg"]
                                ).np_data,
                            },
                        )

                    self._main_model.set_predictions_in_viewer(True)
        # Display popup with saved images path if prediction inputs are from a directory
        else:
            dialog_box = DialogBox(
                f"Predicted images saved to {str(output_path)}. \nWould you like to open this folder?"
            )
            dialog_box.exec()
            if dialog_box.get_selection() and output_path is not None:
                FileUtils.open_directory_in_window(output_path)

        self._main_model.dispatch_prediction_complete()

    def focus_changed(self) -> None:
        self._viewer.clear_layers()



import pytest

from allencell_ml_segmenter._tests.fakes.fake_subscriber import FakeSubscriber
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.thresholding.thresholding_model import (
    ThresholdingModel,
)


@pytest.fixture
def thresholding_model() -> ThresholdingModel:
    model = ThresholdingModel()
    return model


def test_set_thresholding_value_dispatches_event(thresholding_model):
    fake_subscriber: FakeSubscriber = FakeSubscriber()
    thresholding_model.subscribe(
        Event.ACTION_EXECUTE_THRESHOLDING,
        fake_subscriber,
        fake_subscriber.handle,
    )

    thresholding_model.set_thresholding_value(2)

    assert fake_subscriber.was_handled(Event.ACTION_EXECUTE_THRESHOLDING)


def test_set_autothresholding_enabled_dispatches_event(thresholding_model):
    fake_subscriber: FakeSubscriber = FakeSubscriber()
    thresholding_model.subscribe(
        Event.ACTION_EXECUTE_THRESHOLDING,
        fake_subscriber,
        fake_subscriber.handle,
    )

    thresholding_model.set_autothresholding_enabled(True)

    assert fake_subscriber.was_handled(Event.ACTION_EXECUTE_THRESHOLDING)


def test_dispatch_save_thresholded_images(thresholding_model):
    fake_subscriber: FakeSubscriber = FakeSubscriber()
    thresholding_model.subscribe(
        Event.ACTION_SAVE_THRESHOLDING_IMAGES,
        fake_subscriber,
        fake_subscriber.handle,
    )

    thresholding_model.dispatch_save_thresholded_images()

    assert fake_subscriber.was_handled(Event.ACTION_SAVE_THRESHOLDING_IMAGES)



from abc import ABC, abstractmethod
from typing import Callable, Optional, Any


class ITaskExecutor(ABC):
    """
    A TaskExecutor will run the tasks provided to exec() at some point (may be sync or async).
    """

    def __init__(self) -> None:
        raise RuntimeError(
            "Cannot initialize new singleton, please use .global_instance() instead"
        )

    @abstractmethod
    def exec(
        self,
        task: Callable[[], Any],
        on_start: Optional[Callable[[], Any]] = None,
        on_finish: Optional[Callable[[], None]] = None,
        on_return: Optional[Callable[[Any], None]] = None,
        on_error: Optional[Callable[[Exception], None]] = None,
    ) -> None:
        """
        Execute the provided task. Note that on_return must take the return type of task as a param.
        :param task: task to execute
        :param on_start: runs upon the task starting
        :param on_finish: runs upon the task finishing
        :param on_return: runs upon the task returning, must take return type of task as its only param
        :param on_error: runs upon the task throwing an Exception, must take an Exception as its only param
        """
        pass

    @classmethod
    @abstractmethod
    def global_instance(cls):  # type: ignore
        pass



from numpy import zeros, ones, ndarray

from allencell_ml_segmenter.utils.image_processing import (
    set_all_nonzero_values_to,
)


def test_set_all_nonzero_values_to():
    # ARRANGE
    array: ndarray = zeros([3, 3, 3])
    array[1, 1, 1] = 255
    array[2, 1, 1] = 3
    array[0, 1, 1] = 230

    # ACT
    array = set_all_nonzero_values_to(array, 2)

    # ASSERT
    assert array.max() == 2
    assert array.shape == (3, 3, 3)
    assert array[1, 1, 1] == 2
    assert array[2, 1, 1] == 2
    assert array[0, 1, 1] == 2
    assert (
        array.sum() == 6
    )  # we set three non-zero values to 2, so we expect 6


def test_set_all_nonzero_values_to_zero_array():
    # ARRANGE
    array: ndarray = zeros([3, 3, 3])

    # ACT
    array = set_all_nonzero_values_to(array, 1)

    # ASSERT
    assert array.max() == 0
    assert array.shape == (3, 3, 3)
    assert (
        array.sum() == 0
    )  # all values in starting array are 0, so nothing should change


def test_set_all_nonzero_values_to_ones_array():
    # ARRANGE
    array: ndarray = ones([3, 3, 3])

    # ACT
    array = set_all_nonzero_values_to(array, 2)

    # ASSERT
    assert array.max() == 2
    assert array.min() == 2
    assert array.shape == (3, 3, 3)
    assert (
        array.sum() == 54
    )  # we set all values in a array of size 27 to 2, so we expect 54



from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from pathlib import Path
from typing import Optional


class ExperimentUtils:
    """
    ExperimentUtils handles experiment navigation.
    """

    @staticmethod
    def get_best_ckpt(user_exp_path: Path, experiment: str) -> Optional[Path]:
        checkpoints_path = user_exp_path / experiment / "checkpoints"
        if not checkpoints_path.exists():
            return None

        files: list[Path] = [
            entry
            for entry in checkpoints_path.iterdir()
            if entry.is_file() and not "last" in entry.name.lower()
        ]
        if not files:
            return None

        files.sort(key=lambda file: file.stat().st_mtime)
        return checkpoints_path / files[-1].name



from allencell_ml_segmenter.core.task_executor import ITaskExecutor
from typing import Callable, Optional, Any


class SynchroTaskExecutor(ITaskExecutor):

    _instance = None

    def exec(
        self,
        task: Callable[[], Any],
        on_start: Optional[Callable[[], Any]] = None,
        on_finish: Optional[Callable[[], None]] = None,
        on_return: Optional[Callable[[Any], None]] = None,
        on_error: Optional[Callable[[Exception], None]] = None,
    ) -> None:
        if on_start is not None:
            on_start()

        try:
            output: Any = task()
        except Exception as e:
            if on_error is not None:
                on_error(e)
            return

        if on_return is not None:
            on_return(output)
        if on_finish is not None:
            on_finish()

    @classmethod
    def global_instance(cls) -> ITaskExecutor:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance



from pathlib import Path
from typing import Optional
import csv

from qtpy.QtCore import QObject, QThread, Signal
from bioio import BioImage


def extract_channels_from_image(img_path: Path) -> int:
    """
    Returns number of channels in the given img_path.
    :param img_path: image to extract channels from
    """
    img_data: BioImage = BioImage(str(img_path))
    if img_data.dims.T > 1:
        raise RuntimeError("Cannot load timeseries images")

    return img_data.dims.C


def get_img_path_from_csv(csv_path: Path, column: str = "raw") -> Path:
    """
    Returns path of an image in the specified column of the csv or throws exception if there is no data in that column.
    :param csv_path: path to a csv with a 'raw' column
    """
    img_path: Optional[str] = None
    with open(csv_path) as csv_file:
        reader: csv.DictReader = csv.DictReader(csv_file)
        img_path = next(reader)[column]  # type: ignore
    if not img_path:
        raise ValueError(f"No valid data at {csv_path}")
    return Path(img_path).resolve()


class ChannelExtractionThread(QThread):
    """
    A ChannelExtractionThread will extract the number of channels from
    the provided image. If the parent thread has not requested an interruption
    during the thread execution, the number of channels will be emitted through
    the channels_ready signal. If the parent thread has requested an interruption,
    the thread will have no side effects.
    """

    channels_ready: Signal = Signal(int)  # num_channels
    task_failed: Signal = Signal(Exception)

    def __init__(self, img_path: Path, parent: Optional[QObject] = None):
        """
        :param img_path: path to image (must exist, otherwise ValueError)
        :param id: id for this thread instance, provided by parent thread
        """
        super().__init__(parent)
        self._img_path: Path = img_path

    # override
    def run(self) -> None:
        # will show up as a pop-up in the UI, does not force napari to quit
        if not self._img_path.exists():
            raise ValueError(f"{self._img_path} does not exist")

        try:
            channels: int = extract_channels_from_image(self._img_path)
        except Exception as ex:
            self.task_failed.emit(ex)
            return  # return instead of reraise to surprss error message in napari console

        curr_thread: Optional[QThread] = QThread.currentThread()
        if (
            curr_thread is not None
            and not curr_thread.isInterruptionRequested()
        ):
            self.channels_ready.emit(channels)



from unittest.mock import patch, Mock
from pathlib import Path

import pytest
from qtpy.QtWidgets import QFileDialog
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer
from allencell_ml_segmenter.core.file_input_widget import (
    FileInputWidget,
)
from allencell_ml_segmenter.prediction.model import (
    PredictionModel,
)

from allencell_ml_segmenter.core.file_input_model import (
    InputMode,
    FileInputModel,
)

MOCK_PATH: str = "/path/to/file"


@pytest.fixture
def file_input_model(qtbot: QtBot) -> FileInputModel:
    return FileInputModel()


@pytest.fixture
def file_input_widget(
    qtbot: QtBot, file_input_model: FileInputModel
) -> FileInputWidget:
    """
    Fixture that creates an instance of ModelInputWidget for testing.
    """
    return FileInputWidget(file_input_model, viewer=FakeViewer(), service=None)


def test_top_radio_button_slot(
    qtbot: QtBot,
    file_input_widget: FileInputWidget,
    file_input_model: FileInputModel,
) -> None:
    """
    Test the _top_radio_button_slot method of FileInputWidget.
    """
    # ARRANGE - explicitly disable file_input_widget._image_list and enable file_input_widget._browse_dir_edit
    file_input_widget._image_list.setEnabled(False)
    file_input_widget._browse_dir_edit.setEnabled(True)

    # ACT
    with qtbot.waitSignal(file_input_widget._radio_on_screen.toggled):
        file_input_widget._radio_on_screen.click()

    # ASSERT - states should have flipped
    assert file_input_widget._image_list.isEnabled()
    assert not file_input_widget._browse_dir_edit.isEnabled()
    assert file_input_model.get_input_mode() == InputMode.FROM_NAPARI_LAYERS


def test_bottom_radio_button_slot(
    qtbot: QtBot,
    file_input_widget: FileInputWidget,
    file_input_model: FileInputModel,
) -> None:
    """
    Test the _bottom_radio_button_slot method of FileInputWidget.
    """
    # ARRANGE - explicitly enable file_input_widget._image_list and disable file_input_widget._browse_dir_edit
    file_input_widget._image_list.setEnabled(True)
    file_input_widget._browse_dir_edit.setEnabled(False)

    # ACT
    with qtbot.waitSignal(file_input_widget._radio_directory.toggled):
        file_input_widget._radio_directory.click()

    # ASSERT - states should have flipped
    assert not file_input_widget._image_list.isEnabled()
    assert file_input_widget._browse_dir_edit.isEnabled()
    assert file_input_model.get_input_mode() == InputMode.FROM_PATH


# decorator used to stub QFileDialog and avoid nested context managers
@patch.multiple(
    QFileDialog,
    exec_=Mock(return_value=QFileDialog.Accepted),
    selectedFiles=Mock(return_value=[MOCK_PATH]),
    getExistingDirectory=Mock(return_value=MOCK_PATH),
)
def test_populate_input_channel_combobox(qtbot: QtBot) -> None:
    # Arrange
    file_input_model: FileInputModel = FileInputModel()
    prediction_file_input: FileInputWidget = FileInputWidget(
        file_input_model, viewer=FakeViewer(), service=None
    )
    file_input_model.set_max_channels(6)

    # Act
    prediction_file_input._populate_input_channel_combobox()

    # Assert
    assert prediction_file_input._channel_select_dropdown.isEnabled()


def test_input_image_paths(file_input_model: FileInputModel) -> None:
    """
    Tests that the input image paths are set and retrieved properly.
    """
    # ARRANGE
    dummy_paths: List[Path] = [
        Path("example path " + str(i)) for i in range(10)
    ]

    # ACT
    file_input_model.set_input_image_path(dummy_paths)

    # ASSERT
    assert file_input_model.get_input_image_path() == dummy_paths


def test_image_input_channel_index(file_input_model: FileInputModel) -> None:
    """
    Tests that the channel index is set and retrieved properly.
    """
    for i in range(10):
        # ACT
        file_input_model.set_image_input_channel_index(i)

        # ASSERT
        assert file_input_model.get_image_input_channel_index() == i


def test_output_directory(file_input_model: FileInputModel) -> None:
    """
    Tests that the output directory is set and retrieved properly.
    """
    # ARRANGE
    dummy_path: Path = Path("example path")

    # ACT
    file_input_model.set_output_directory(dummy_path)

    # ASSERT
    assert file_input_model.get_output_directory() == dummy_path



from typing import Optional, Dict
from allencell_ml_segmenter.config.i_user_settings import IUserSettings

from allencell_ml_segmenter.main.viewer import Viewer

import napari  # type: ignore
from allencell_ml_segmenter.curation.curation_widget import CurationWidget
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QVBoxLayout,
    QSizePolicy,
    QTabWidget,
    QWidget,
)
from allencell_ml_segmenter.config.user_settings import UserSettings

from allencell_ml_segmenter.core.aics_widget import AicsWidget

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.view import MainWindow
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.main.main_service import MainService
from allencell_ml_segmenter.main.i_viewer import IViewer
from allencell_ml_segmenter.prediction.model import PredictionModel
from allencell_ml_segmenter.prediction.view import PredictionView
from allencell_ml_segmenter.services.prediction_service import (
    PredictionService,
)
from allencell_ml_segmenter.services.training_service import TrainingService
from allencell_ml_segmenter.thresholding.thresholding_model import (
    ThresholdingModel,
)
from allencell_ml_segmenter.thresholding.thresholding_service import (
    ThresholdingService,
)
from allencell_ml_segmenter.training.model_selection_widget import (
    ModelSelectionWidget,
)
from allencell_ml_segmenter.training.training_model import TrainingModel
from allencell_ml_segmenter.training.view import TrainingView
from allencell_ml_segmenter.curation.curation_model import CurationModel
from allencell_ml_segmenter._style import Style
from allencell_ml_segmenter.curation.curation_service import CurationService
from allencell_ml_segmenter.thresholding.thresholding_view import (
    ThresholdingView,
)
from allencell_ml_segmenter.core.file_input_model import FileInputModel


class MainWidget(AicsWidget):
    """
    Holds the pertinent view at the moment to be displayed to the user.
    """

    def __init__(
        self, viewer: napari.Viewer, settings: Optional[IUserSettings] = None
    ):
        super().__init__()
        self.user_settings: IUserSettings
        if settings is None:
            self.user_settings = UserSettings()
        else:
            self.user_settings = settings

        self.viewer: IViewer = Viewer(viewer)

        # basic styling
        self.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding
        )
        layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)

        self._model: MainModel = MainModel()
        self._model.subscribe(
            Event.ACTION_CHANGE_VIEW, self, self._handle_change_view
        )
        self._model.subscribe(
            Event.ACTION_NEW_MODEL, self, self._handle_new_model
        )

        if self.user_settings.get_user_experiments_path() is None:
            self.user_settings.prompt_for_user_experiments_home(parent=self)

        # init models
        self._experiments_model = ExperimentsModel(self.user_settings)

        self._training_model: TrainingModel = TrainingModel(
            main_model=self._model, experiments_model=self._experiments_model
        )
        self._prediction_file_input_model: FileInputModel = FileInputModel()
        self._prediction_model: PredictionModel = PredictionModel()
        self._curation_model: CurationModel = CurationModel(
            self._experiments_model,
            self._model,
        )

        self._thresholding_model: ThresholdingModel = ThresholdingModel()

        # init services
        self._main_service: MainService = MainService(
            self._model, self._experiments_model
        )
        self._curation_service: CurationService = CurationService(
            self._curation_model,
            self._experiments_model,
        )
        self._training_service: TrainingService = TrainingService(
            training_model=self._training_model,
            experiments_model=self._experiments_model,
        )
        self._prediction_service: PredictionService = PredictionService(
            prediction_model=self._prediction_model,
            file_input_model=self._prediction_file_input_model,
            experiments_model=self._experiments_model,
        )

        self._thresholding_service: ThresholdingService = ThresholdingService(
            thresholding_model=self._thresholding_model,
            experiments_model=self._experiments_model,
            main_model=self._model,
            viewer=self.viewer,
        )

        # keep track of windows
        self._window_container: QTabWidget = QTabWidget()
        self._window_container.setDisabled(True)
        self._window_to_index: dict[MainWindow, int] = dict()

        self._experiments_model.subscribe(
            Event.ACTION_EXPERIMENT_APPLIED,
            self,
            self._handle_experiment_applied,
        )

        # initialize the tabs
        self._curation_view: CurationWidget = CurationWidget(
            self.viewer, self._curation_model
        )
        self._initialize_window(self._curation_view, "Curation")
        self._training_view: TrainingView = TrainingView(
            main_model=self._model,
            viewer=self.viewer,
            experiments_model=self._experiments_model,
            training_model=self._training_model,
        )
        self._initialize_window(self._training_view, "Training")
        self._prediction_view: PredictionView = PredictionView(
            main_model=self._model,
            prediction_model=self._prediction_model,
            file_input_model=self._prediction_file_input_model,
            viewer=self.viewer,
        )
        self._initialize_window(self._prediction_view, "Prediction")
        self._window_container.currentChanged.connect(self._tab_changed)

        self._thresholding_view = ThresholdingView(
            main_model=self._model,
            thresholding_model=self._thresholding_model,
            experiments_model=self._experiments_model,
            viewer=self.viewer,
        )
        self._initialize_window(self._thresholding_view, "Thresholding")

        # Model selection which applies to all views
        model_selection_widget: ModelSelectionWidget = ModelSelectionWidget(
            self._model, self._experiments_model, self.user_settings
        )
        model_selection_widget.setObjectName("modelSelection")

        layout.addWidget(model_selection_widget, Qt.AlignmentFlag.AlignTop)
        layout.addWidget(self._window_container, Qt.AlignmentFlag.AlignCenter)
        layout.addStretch(100)
        self.setStyleSheet(Style.get_stylesheet("core.qss"))

        # events for auto window switching
        self._model.subscribe(
            Event.PROCESS_TRAINING_COMPLETE,
            self,
            self._disable_non_prediction_tabs,
        )

    def _handle_experiment_applied(self, _: Event) -> None:
        """
        Handle the experiment applied event.
        """
        self._window_container.setDisabled(
            self._experiments_model.get_experiment_name() is None
        )

    def _handle_new_model(self, _: Event) -> None:
        """
        Handle the new model radio button toggled event.

        inputs:
            is_new_model - bool
        """
        self._window_container.setTabEnabled(0, self._model.is_new_model())
        self._window_container.setTabEnabled(1, self._model.is_new_model())
        self._set_window(
            self._curation_view
            if self._model.is_new_model()
            else self._prediction_view
        )
        self._window_container.setTabEnabled(3, True)  # TODO remove

    def _disable_non_prediction_tabs(self, _: Event) -> None:
        """
        Handle existing model selection (disable tabs).

        inputs:
            is_new_model - bool
        """
        self._window_container.setTabEnabled(0, False)
        self._window_container.setTabEnabled(1, False)
        self._window_container.setTabEnabled(3, True)  # TODO remove

    def _handle_change_view(self, event: Event) -> None:
        """
        Handle event function for the main widget, which handles MainEvents.

        inputs:
            event - MainEvent
        """
        curr_view: Optional[MainWindow] = self._model.get_current_view()
        if curr_view is not None:
            self._set_window(curr_view)

    def _set_window(self, window: MainWindow) -> None:
        """
        Set the current Window (prediction/training/curation), must be initialized first
        """
        self._window_container.setCurrentIndex(self._window_to_index[window])

    def _initialize_window(self, window: MainWindow, title: str) -> None:
        # QTabWidget count method keeps track of how many child widgets have been added
        self._window_to_index[window] = self._window_container.count()
        self._window_container.addTab(window, title)  # type: ignore

    def _tab_changed(self, index: int) -> None:
        """
        Resize bottom edge of QTabWidget to fit the current view.
        """
        for i in range(self._window_container.count()):
            widget: Optional[QWidget] = self._window_container.widget(i)
            if widget is not None and i == index:
                widget.setSizePolicy(
                    QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum
                )
            elif widget is not None:
                widget.setSizePolicy(
                    QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Ignored
                )

                # call the focus_changed method of the view that we are switching to
        list(self._window_to_index.keys())[index].focus_changed()



import platform
from typing import Dict, Union, Optional, List

from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ImageType,
    ModelSize,
)
from allencell_ml_segmenter.utils.cuda_util import CUDAUtils


class CytoDLOverridesManager:
    """
    Class to generate overrides for cyto-dl based on user selections in app state
    """

    # TODO use a similar pattern for Predictions- willdo in another ticket
    # https://github.com/AllenCell/allencell-ml-segmenter/issues/234
    def __init__(
        self,
        experiments_model: ExperimentsModel,
        training_model: Optional[TrainingModel] = None,
    ) -> None:
        self._experiments_model: ExperimentsModel = experiments_model
        self._training_model: Optional[TrainingModel] = training_model

    def get_training_overrides(
        self,
    ) -> Dict[str, Union[str, int, float, bool, Dict, List]]:
        # check to see if CytoOverridesManager was constructed with a training model
        if self._training_model is None:
            raise ValueError(
                "CytoOverridesManager must be constructed with a training model in order to get training overrides."
            )

        overrides_dict: Dict[str, Union[str, int, float, bool, Dict, List]] = (
            dict()
        )

        # ITERATIVE TRAINING
        # if pulling weights from an existing model
        if self._training_model.is_using_existing_model():
            # use best checkpoint from existing model to pull weights from
            overrides_dict["checkpoint.ckpt_path"] = str(
                self._training_model.get_existing_model_ckpt_path()
            )
            # needed for pulling weights
            overrides_dict["checkpoint.weights_only"] = True
            overrides_dict["checkpoint.strict"] = False
            # ensure correct output path for these models
            overrides_dict["paths.output_dir"] = (
                f"{self._experiments_model.get_user_experiments_path()}/{self._experiments_model.get_experiment_name()}"
            )

        else:
            model_size: Optional[ModelSize] = (
                self._training_model.get_model_size()
            )
            if model_size is None:
                raise ValueError(
                    "Model size is required, but is not set, and get_training_overrides was called."
                )
            # Filters/Model Size (required if starting new model)
            overrides_dict["model._aux.filters"] = model_size.value

        # Hardware overrides (required)
        if CUDAUtils.cuda_available():
            overrides_dict["trainer.accelerator"] = "gpu"
        else:
            overrides_dict["trainer.accelerator"] = "cpu"
        overrides_dict["data.num_workers"] = CUDAUtils.get_num_workers()

        # pytorch needs num_workers = 0 for mac osx
        # keeps prediction on main thread
        if platform.system() == "Darwin":
            overrides_dict["data.num_workers"] = 0

        # Spatial Dims (required)
        dims: Optional[int] = self._training_model.get_spatial_dims()
        if dims is not None:
            overrides_dict["spatial_dims"] = dims

        # Max Run
        # define max run (in epochs, required)
        current_epoch: Optional[int] = (
            self._experiments_model.get_current_epoch()
        )
        num_epochs: Optional[int] = self._training_model.get_num_epochs()
        if num_epochs is not None:
            adjusted_epoch: int = (
                current_epoch + 1 if current_epoch is not None else 0
            )
            overrides_dict["trainer.max_epochs"] = num_epochs + adjusted_epoch
        # max run in time is also defined (optional)
        if self._training_model.use_max_time():
            # define max runtime (in hours)
            overrides_dict["trainer.max_time"] = {
                "minutes": self._training_model.get_max_time()
            }

        # Training input path (required)
        if self._training_model.get_images_directory() is None:
            raise ValueError(
                "Training data path was not set but get_training_overrides was called."
            )
        overrides_dict["data.path"] = str(
            self._training_model.get_images_directory()
        )

        # patch size (required)
        patch_size: Optional[List[int]] = self._training_model.get_patch_size()
        if patch_size is None:
            raise ValueError(
                "Patch size is required, but is not set, and get_training_overrides was called."
            )
        overrides_dict["data._aux.patch_shape"] = patch_size

        # Commented out 5/23 brian.kim
        # We no longer support training from a old checkpoint- leaving this in if we want to re-enable this in the
        # future to continue training from an existing checkpoint.
        # if self._experiments_model.get_checkpoint() is not None:
        #     # We are going to continue training on an existing model
        #     overrides_dict["ckpt_path"] = str(
        #         self._experiments_model.get_model_checkpoints_path(
        #             self._experiments_model.get_experiment_name(),
        #             self._experiments_model.get_checkpoint(),
        #         )
        #     )
        # Channel Override
        raw_channel: Optional[int] = self._training_model.get_selected_channel(
            ImageType.RAW
        )
        seg1_channel: Optional[int] = (
            self._training_model.get_selected_channel(ImageType.SEG1)
        )
        seg2_channel: Optional[int] = (
            self._training_model.get_selected_channel(ImageType.SEG2)
        )
        if raw_channel is not None:
            overrides_dict["input_channel"] = raw_channel
        if seg1_channel is not None:
            overrides_dict["target_col1_channel"] = seg1_channel
        if seg2_channel is not None:
            overrides_dict["target_col2_channel"] = seg2_channel
        return overrides_dict



from pathlib import Path
from typing import Optional

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QSizePolicy,
    QFrame,
    QGridLayout,
    QComboBox,
)

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.curation.stacked_spinner import StackedSpinner
from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ImageType,
)
from allencell_ml_segmenter.widgets.input_button_widget import (
    InputButton,
    FileInputMode,
)
from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint


class ImageSelectionWidget(QWidget):
    """
    A widget for training image selection.
    """

    TITLE_TEXT: str = "Training images"

    def __init__(
        self, model: TrainingModel, experiments_model: IExperimentsModel
    ):
        super().__init__()

        self._model: TrainingModel = model
        self._experiments_model: IExperimentsModel = experiments_model

        # widget skeleton
        layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum
        )

        title: LabelWithHint = LabelWithHint(ImageSelectionWidget.TITLE_TEXT)
        # TODO: hints for widget titles?
        title.setObjectName("title")
        layout.addWidget(title)

        frame: QFrame = QFrame()
        frame_layout: QGridLayout = QGridLayout()
        frame.setLayout(frame_layout)
        frame_layout.setSpacing(0)
        frame.setObjectName("frame")
        layout.addWidget(frame)

        # grid contents
        directory_label: LabelWithHint = LabelWithHint(
            "Curated image data source"
        )
        directory_label.set_hint(
            "The image dataset output by Curation. Directory should contain train.csv, val.csv, and test.csv files."
        )
        self._images_directory_input_button: InputButton = InputButton(
            self._model,
            lambda dir: self._on_input_images_select(dir),
            "Select directory...",
            FileInputMode.DIRECTORY_OR_CSV,
        )
        self._images_directory_input_button.elongate(248)
        self._training_data_stacked_spinner = StackedSpinner(
            self._images_directory_input_button
        )

        frame_layout.setSpacing(10)
        frame_layout.addWidget(
            directory_label, 0, 0, Qt.AlignmentFlag.AlignVCenter
        )

        # remove until we provide tutorial for manually creating csv
        # guide_text: QLabel = QLabel()
        # guide_text.setText(
        #    "<a href='https://www.allencell.org/segmenter.html'>See instructions</a> for CSV file"
        # )
        # guide_text.setObjectName("guideText")
        # guide_text.setTextFormat(Qt.RichText)
        # guide_text.setOpenExternalLinks(True)

        frame_layout.addWidget(
            self._training_data_stacked_spinner,
            0,
            1,
            Qt.AlignmentFlag.AlignVCenter,
        )
        # frame_layout.addWidget(guide_text, 1, 1, Qt.AlignTop)

        self._raw_channel_combo_box: QComboBox = QComboBox()
        self._raw_channel_combo_box.setMinimumWidth(306)
        self._reset_combo_box(self._raw_channel_combo_box, None)
        self._raw_channel_combo_box.currentIndexChanged.connect(
            lambda idx: self._handle_idx_change(idx, ImageType.RAW)
        )
        frame_layout.addWidget(LabelWithHint("Raw image channel"), 2, 0)
        frame_layout.addWidget(self._raw_channel_combo_box, 2, 1)

        self._seg1_channel_combo_box: QComboBox = QComboBox()
        self._seg1_channel_combo_box.setMinimumWidth(306)
        self._reset_combo_box(self._seg1_channel_combo_box, None)
        self._seg1_channel_combo_box.currentIndexChanged.connect(
            lambda idx: self._handle_idx_change(idx, ImageType.SEG1)
        )
        frame_layout.addWidget(LabelWithHint("Seg1 image channel"), 3, 0)
        frame_layout.addWidget(self._seg1_channel_combo_box, 3, 1)

        self._seg2_channel_combo_box: QComboBox = QComboBox()
        self._seg2_channel_combo_box.setMinimumWidth(306)
        self._reset_combo_box(self._seg2_channel_combo_box, None)
        self._seg2_channel_combo_box.currentIndexChanged.connect(
            lambda idx: self._handle_idx_change(idx, ImageType.SEG2)
        )
        frame_layout.addWidget(LabelWithHint("Seg2 image channel"), 4, 0)
        frame_layout.addWidget(self._seg2_channel_combo_box, 4, 1)

        self._experiments_model.subscribe(
            Event.ACTION_EXPERIMENT_APPLIED, self, self.set_inputs_csv
        )

        self._model.signals.num_channels_set.connect(self._update_channels)

    def set_inputs_csv(self, event: Optional[Event] = None) -> None:
        csv_path: Optional[Path] = self._experiments_model.get_csv_path()
        if csv_path is not None:
            csv_path = csv_path / "train.csv"
            if csv_path.is_file():
                # if the csv exists
                self._images_directory_input_button._text_display.setText(
                    str(self._experiments_model.get_csv_path())
                )
                # This also dispatches channel extraction
                self._model.set_images_directory(
                    self._experiments_model.get_csv_path()
                )
            else:
                self._images_directory_input_button._text_display.setText("")
                self._model.set_images_directory(None)

    def _on_input_images_select(self, dir: Path) -> None:
        self._set_to_loading()
        # This also dispatches channel extraction
        self._model.set_images_directory(dir)

    def _set_to_loading(self) -> None:
        self._training_data_stacked_spinner.start()
        self._reset_combo_box(self._raw_channel_combo_box, None)
        self._reset_combo_box(self._seg1_channel_combo_box, None)
        self._reset_combo_box(self._seg2_channel_combo_box, None)

    def _handle_idx_change(self, idx: int, img_type: ImageType) -> None:
        self._model.set_selected_channel(img_type, idx if idx >= 0 else None)

    def _reset_combo_box(
        self,
        combo_box: QComboBox,
        num_channels: Optional[int],
        default_channel: Optional[int] = None,
    ) -> None:
        if default_channel is None:
            default_channel = 0

        combo_box.clear()
        if num_channels is not None:
            combo_box.addItems([str(x) for x in range(num_channels)])
            combo_box.setCurrentIndex(
                default_channel if default_channel < num_channels else 0
            )
            combo_box.setEnabled(True)
        else:
            combo_box.setPlaceholderText("")
            combo_box.setCurrentIndex(-1)
            combo_box.setEnabled(False)

    def _update_channels(self) -> None:
        self._training_data_stacked_spinner.stop()
        default_channels: dict[ImageType, Optional[int]] = (
            self._model.get_selected_channels()
        )

        self._reset_combo_box(
            self._raw_channel_combo_box,
            self._model.get_num_channels(ImageType.RAW),
            default_channel=default_channels[ImageType.RAW],
        )
        self._reset_combo_box(
            self._seg1_channel_combo_box,
            self._model.get_num_channels(ImageType.SEG1),
            default_channel=default_channels[ImageType.SEG1],
        )
        self._reset_combo_box(
            self._seg2_channel_combo_box,
            self._model.get_num_channels(ImageType.SEG2),
            default_channel=default_channels[ImageType.SEG2],
        )



from enum import Enum


class Event(Enum):
    """
    Different application Events
    """

    # Process events. This signals that a long-running process is active. Progress updates should be shown to the user.
    PROCESS_TRAINING = "training"
    PROCESS_TRAINING_PROGRESS = "training_progress"
    PROCESS_TRAINING_SHOW_ERROR = "training_error"
    PROCESS_TRAINING_CLEAR_ERROR = "training_clear_error"
    PROCESS_TRAINING_COMPLETE = "training_complete"
    PROCESS_PREDICTION = "prediction"
    PROCESS_PREDICTION_COMPLETE = "prediction_complete"
    PROCESS_CURATION_INPUT_STARTED = "curation_input_started"
    PROCESS_CURATION_NEXT_IMAGE = "curation_next_image"

    # Action events. This signals a change in the UI. These are a direct result of a user action
    ACTION_REFRESH = "refresh"
    ACTION_CHANGE_VIEW = "change_view"
    ACTION_NEW_MODEL = "new_model"
    # Training
    ACTION_START_TRAINING = "start_training"
    ACTION_TRAINING_MAX_NUMBER_CHANNELS_SET = "max_number_channel_set_training"
    ACTION_TRAINING_DATASET_SELECTED = "training_dataset_selected"
    # Prediction
    ACTION_PREDICTION_MODEL_FILE = "model_file"
    ACTION_PREDICTION_PREPROCESSING_METHOD = "preprocessing_method"
    ACTION_PREDICTION_POSTPROCESSING_METHOD = "postprocessing_method"
    ACTION_PREDICTION_POSTPROCESSING_SIMPLE_THRESHOLD = (
        "postprocessing_simple_threshold"
    )
    ACTION_PREDICTION_POSTPROCESSING_AUTO_THRESHOLD = (
        "postprocessing_auto_threshold"
    )
    ACTION_PREDICTION_GET_IMAGE_PATHS_FROM_NAPARI = (
        "prediction_get_image_paths_from_napari"
    )

    ACTION_PREDICTION_SETUP = "prediction_setup"

    # FILEINPUT
    ACTION_FILEINPUT_MAX_CHANNELS_SET = "fileinput_max_channels_set"
    ACTION_FILEINPUT_EXTRACT_CHANNELS = "fileinput_extract_channels"

    # Curation
    ACTION_CURATION_RAW_CHANNELS_SET = "curation_raw_channels_set"
    ACTION_CURATION_SEG1_CHANNELS_SET = "curation_seg1_channels_set"
    ACTION_CURATION_SEG2_CHANNELS_SET = "curation_seg2_channels_set"
    # Experiment related
    ACTION_EXPERIMENT_SELECTED = "experiment_selected"
    ACTION_EXPERIMENT_APPLIED = "experiment_applied"
    ACTION_CURATION_DRAW_EXCLUDING = "curation_draw_excluding"
    ACTION_CURATION_FINISHED_DRAW_EXCLUDING = (
        "curation_finished_draw_excluding"
    )
    ACTION_CURATION_DRAW_MERGING = "curation_draw_merging"
    ACTION_CURATION_SAVE_EXCLUDING_MASK = "curation_save_excluding_mask"
    ACTION_CURATION_SAVED_MERGING_MASK = "curation_saved_merging_mask"
    # error handling
    ACTION_CURATION_SEG2_THREAD_ERROR = "curation_seg2_thread_error"
    ACTION_CURATION_SEG1_THREAD_ERROR = "curation_seg1_thread_error"
    ACTION_CURATION_RAW_THREAD_ERROR = "curation_raw_thread_error"

    # Thresholding events
    ACTION_EXECUTE_THRESHOLDING = "execute_thresholding"
    ACTION_SAVE_THRESHOLDING_IMAGES = "save_thresholding_images"
    ACTION_THRESHOLDING_DISABLED = "disable_thresholding"

    # View selection events. These can stem from a user action, or from a process (i.e. prediction process ends, and a new view is shown automatically).
    VIEW_SELECTION_TRAINING = "training_selected"
    VIEW_SELECTION_PREDICTION = "prediction_selected"



from pathlib import Path

import allencell_ml_segmenter
from allencell_ml_segmenter.core.directories import Directories


def test_get_assets_dir():
    """
    Test that the assets directory is returned correctly.
    """
    # ARRANGE
    assets_path: Path = Path(allencell_ml_segmenter.__file__).parent / "assets"

    # ACT/ASSERT
    assert Directories.get_assets_dir() == assets_path


def test_get_style_dir():
    """
    Test that the styles directory is returned correctly.
    """
    # ARRANGE
    styles_path: Path = Path(allencell_ml_segmenter.__file__).parent / "styles"

    # ACT/ASSERT
    assert Directories.get_style_dir() == styles_path



from pathlib import Path

import pytest
from unittest.mock import patch

from qtpy.QtWidgets import QFileDialog
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter.prediction.model_input_widget import (
    ModelInputWidget,
)
from allencell_ml_segmenter.prediction.model import PredictionModel
from allencell_ml_segmenter.core.file_input_model import FileInputModel


@pytest.fixture
def model_input_widget(qtbot: QtBot):
    """
    Fixture that creates an instance of ModelInputWidget for testing.
    """
    return ModelInputWidget(PredictionModel())


def test_top_radio_button_slot(
    qtbot: QtBot, model_input_widget: ModelInputWidget
) -> None:
    """
    Test the _top_radio_button_slot method of ModelInputWidget.
    """
    # ARRANGE - explicitly enable all input UI elements (slider and combo box)
    model_input_widget._simple_thresh_slider.native.setEnabled(True)
    model_input_widget._auto_thresh_selection.setEnabled(True)

    # ACT
    with qtbot.waitSignal(model_input_widget._top_postproc_button.toggled):
        model_input_widget._top_postproc_button.click()

    # ASSERT
    assert not model_input_widget._simple_thresh_slider.native.isEnabled()
    assert not model_input_widget._auto_thresh_selection.isEnabled()


def test_mid_radio_button_slot(
    qtbot: QtBot, model_input_widget: ModelInputWidget
) -> None:
    """
    Test the _mid_radio_button_slot method of ModelInputWidget.
    """
    # ARRANGE - explicitly disable the slider and enable the combo box
    model_input_widget._simple_thresh_slider.native.setEnabled(False)
    model_input_widget._auto_thresh_selection.setEnabled(True)

    # ACT
    with qtbot.waitSignal(model_input_widget._mid_postproc_button.toggled):
        model_input_widget._mid_postproc_button.click()

    # ASSERT - slider should be enabled, combo box should be disabled, postprocessing method should be set to MID_TEXT
    assert model_input_widget._simple_thresh_slider.native.isEnabled()
    assert not model_input_widget._auto_thresh_selection.isEnabled()


def test_bottom_radio_button_slot(
    qtbot: QtBot, model_input_widget: ModelInputWidget
) -> None:
    """
    Test the _bottom_radio_button_slot method of ModelInputWidget.
    """
    # ARRANGE - explicitly enable the slider and disable the combo box
    model_input_widget._simple_thresh_slider.native.setEnabled(True)
    model_input_widget._auto_thresh_selection.setEnabled(False)

    # ACT
    with qtbot.waitSignal(model_input_widget._bottom_postproc_button.toggled):
        model_input_widget._bottom_postproc_button.click()

    # ASSERT - slider should be disabled, combo box should be enabled, postprocessing method should be set to BOTTOM_TEXT
    assert not model_input_widget._simple_thresh_slider.native.isEnabled()
    assert model_input_widget._auto_thresh_selection.isEnabled()


def test_postprocessing_method(
    qtbot: QtBot, model_input_widget: ModelInputWidget
) -> None:
    """
    Tests that selecting the associated radio buttons updates the postprocessing method in the model.
    """
    # ACT
    with qtbot.waitSignal(model_input_widget._top_postproc_button.toggled):
        model_input_widget._top_postproc_button.click()

    # ACT
    with qtbot.waitSignal(model_input_widget._mid_postproc_button.toggled):
        model_input_widget._mid_postproc_button.click()

    # ASSERT

    # ACT
    with qtbot.waitSignal(model_input_widget._bottom_postproc_button.toggled):
        model_input_widget._bottom_postproc_button.click()



from abc import ABC, abstractmethod
import numpy as np
from pathlib import Path
from typing import List, Union


class IFileWriter(ABC):
    def __init__(self) -> None:
        raise RuntimeError(
            "Cannot initialize new singleton, please use .global_instance() instead"
        )

    @abstractmethod
    def np_save(self, path: Path, arr: np.ndarray) -> None:
        """
        Saves :param arr: to :param path:. Creates directories as necessary.
        """
        pass

    @abstractmethod
    def csv_open_write_mode(self, path: Path) -> None:
        """
        Opens a CSV at :param path: in write mode. Can only call csv_write_row
        on the path after running this method. Must call csv_close with the same
        path at some point after calling this method. Creates directories as necessary.
        """
        pass

    @abstractmethod
    def csv_write_row(self, path: Path, row: List[str]) -> None:
        """
        Appends :param row: to the open CSV at :param path:
        """
        pass

    @abstractmethod
    def csv_close(self, path: Path) -> None:
        """
        Closes the open CSV at :param path:
        """
        pass

    @abstractmethod
    def write_json(self, json_like_obj: Union[list, dict], path: Path) -> None:
        pass



from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.main.main_service import MainService
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ModelSize,
    ImageType,
)
from allencell_ml_segmenter.training.view import TrainingView
from allencell_ml_segmenter.core.task_executor import SynchroTaskExecutor
import pytest
from pytestqt.qtbot import QtBot
import allencell_ml_segmenter
from pathlib import Path


@pytest.fixture
def main_model():
    return MainModel()


@pytest.fixture
def experiments_model():
    return FakeExperimentsModel()


@pytest.fixture
def training_model(main_model, experiments_model):
    return TrainingModel(
        main_model=main_model, experiments_model=experiments_model
    )


@pytest.fixture
def viewer():
    return FakeViewer()


@pytest.fixture
def training_view(
    qtbot: QtBot, main_model: MainModel, training_model: TrainingModel
) -> TrainingView:
    """
    Returns a PredictionView instance for testing.
    """
    experimentsModel = FakeExperimentsModel()
    return TrainingView(
        main_model=main_model,
        experiments_model=experimentsModel,
        training_model=training_model,
        viewer=FakeViewer(),
    )


def test_set_image_dimensions(
    qtbot: QtBot,
    training_view: TrainingView,
    training_model: TrainingModel,
) -> None:
    """
    Tests that checking the associated radio buttons properly sets the image dimensions.
    """
    # ASSERT (initial state)
    assert not training_view.x_patch_size.isEnabled()
    assert not training_view.y_patch_size.isEnabled()
    assert not training_view.z_patch_size.isEnabled()

    # ACT
    with qtbot.waitSignal(training_model.signals.spatial_dims_set):
        training_model.set_spatial_dims(2)

    # ASSERT
    assert training_view.x_patch_size.isEnabled()
    assert training_view.y_patch_size.isEnabled()
    assert not training_view.z_patch_size.isEnabled()

    # ACT
    with qtbot.waitSignal(training_model.signals.spatial_dims_set):
        training_model.set_spatial_dims(3)

    # ASSERT
    assert training_view.x_patch_size.isEnabled()
    assert training_view.y_patch_size.isEnabled()
    assert training_view.z_patch_size.isEnabled()


def test_set_max_epoch(
    qtbot: QtBot, training_view: TrainingView, training_model: TrainingModel
) -> None:
    """
    Tests that the max epoch field is properly set by the associated QLineEdit.
    """
    # ACT
    qtbot.keyClicks(training_view._num_epochs_input, "100")

    # ASSERT
    assert training_model.get_num_epochs() == 100


def test_set_model_size(
    training_view: TrainingView, training_model: TrainingModel
) -> None:
    """
    Tests that using the associated combo box properly sets the model size field.
    """
    for idx, model_size in enumerate(ModelSize):
        # ACT
        training_view._model_size_combo_box.setCurrentIndex(idx)

        # ASSERT
        assert training_model.get_model_size() == model_size


def test_set_patch_size(
    qtbot: QtBot,
    main_model: MainModel,
    experiments_model: FakeExperimentsModel,
    training_model: TrainingModel,
) -> None:
    # ARRANGE
    view: TrainingView = TrainingView(
        main_model=main_model,
        experiments_model=FakeExperimentsModel(),
        training_model=training_model,
        viewer=FakeViewer(),
    )
    view.z_patch_size.setText("1")
    view.y_patch_size.setText("4")
    view.x_patch_size.setText("12")

    # ACT
    view.set_patch_size()

    # ASSERT
    assert training_model.get_patch_size() == [1, 4, 12]


def test_navigate_to_training_populates_channel_selection(
    qtbot: QtBot,
) -> None:
    """
    Validates that when there is a valid channel selection JSON, navigating to training
    will auto-populate the channel selection dropdowns with the previously chosen channels
    from curation.
    """
    # Arrange
    test_channel_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "channel_selection_json"
        / "valid_mixed.json"
    )
    main_model: MainModel = MainModel()
    experiments_model: FakeExperimentsModel = FakeExperimentsModel(
        channel_selection_path=test_channel_path
    )
    # must init main service and set experiment name to pull channel data into main model
    main_service: MainService = MainService(
        main_model,
        experiments_model,
        task_executor=SynchroTaskExecutor.global_instance(),
    )
    experiments_model.apply_experiment_name("test")

    training_model: TrainingModel = TrainingModel(
        main_model, experiments_model
    )
    view: TrainingView = TrainingView(
        main_model=main_model,
        experiments_model=experiments_model,
        training_model=training_model,
        viewer=FakeViewer(),
    )

    # Act
    # simulate service completing its channel extraction work
    training_model.set_all_num_channels(
        {
            ImageType.RAW: 8,
            ImageType.SEG1: 6,
            ImageType.SEG2: 4,
        }
    )

    # Assert (these values come from valid_mixed.json)
    assert training_model.get_selected_channel(ImageType.RAW) == 5
    assert training_model.get_selected_channel(ImageType.SEG1) == 2
    assert training_model.get_selected_channel(ImageType.SEG2) == 1
    assert (
        view.image_selection_widget._raw_channel_combo_box.currentIndex() == 5
    )
    assert (
        view.image_selection_widget._seg1_channel_combo_box.currentIndex() == 2
    )
    assert (
        view.image_selection_widget._seg2_channel_combo_box.currentIndex() == 1
    )


def test_navigate_to_training_populates_channel_selection_no_json(
    qtbot: QtBot,
) -> None:
    """
    Validates that when there is no channel selection JSON, navigating to training
    will auto-populate the channel selection dropdowns with 0.
    """
    # Arrange
    test_channel_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "channel_selection_json"
        / "nonexistent.json"
    )
    main_model: MainModel = MainModel()
    experiments_model: FakeExperimentsModel = FakeExperimentsModel(
        channel_selection_path=test_channel_path
    )
    # must init main service and set experiment to pull channel data into main model
    main_service: MainService = MainService(
        main_model,
        experiments_model,
        task_executor=SynchroTaskExecutor.global_instance(),
    )
    experiments_model.apply_experiment_name("test")

    training_model: TrainingModel = TrainingModel(
        main_model, experiments_model
    )
    view: TrainingView = TrainingView(
        main_model=main_model,
        experiments_model=experiments_model,
        training_model=training_model,
        viewer=FakeViewer(),
    )

    # Act
    # simulate service completing its channel extraction work
    training_model.set_all_num_channels(
        {
            ImageType.RAW: 8,
            ImageType.SEG1: 6,
            ImageType.SEG2: 4,
        }
    )

    # Assert (these values come from valid_mixed.json)
    assert training_model.get_selected_channel(ImageType.RAW) == 0
    assert training_model.get_selected_channel(ImageType.SEG1) == 0
    assert training_model.get_selected_channel(ImageType.SEG2) == 0
    assert (
        view.image_selection_widget._raw_channel_combo_box.currentIndex() == 0
    )
    assert (
        view.image_selection_widget._seg1_channel_combo_box.currentIndex() == 0
    )
    assert (
        view.image_selection_widget._seg2_channel_combo_box.currentIndex() == 0
    )


def test_existing_model_radio(
    qtbot: QtBot, training_view: TrainingView, training_model: TrainingModel
) -> None:
    """
    Test the slots connected to the existing model radio selection
    """
    # Some checks before testing
    assert (
        not training_view.existing_model_dropdown.isEnabled()
    )  # model dropdown is disabled by default
    assert (
        training_view.existing_model_no_radio.isChecked()
    )  # no radio is checked by default
    assert not training_view.existing_model_yes_radio.isChecked()

    # ACT (click yes radio)
    with qtbot.waitSignal(training_view.existing_model_yes_radio.clicked):
        training_view.existing_model_yes_radio.click()

    # ASSERT
    # radio buttons flipped
    assert not training_view.existing_model_no_radio.isChecked()
    assert training_view.existing_model_yes_radio.isChecked()
    # model dropdown is enabled
    assert training_view.existing_model_dropdown.isEnabled()
    # training model updated
    assert training_model.is_using_existing_model()

    # ACT (click no radio again)
    with qtbot.waitSignal(training_view.existing_model_no_radio.clicked):
        training_view.existing_model_no_radio.click()

    # ASSERT
    # radio buttons flipped
    assert training_view.existing_model_no_radio.isChecked()
    assert not training_view.existing_model_yes_radio.isChecked()
    # model dropdown is disabled
    assert not training_view.existing_model_dropdown.isEnabled()
    # training model updated
    assert not training_model.is_using_existing_model()
    assert training_model.get_existing_model() is None


def test_model_size_combo_box(
    qtbot: QtBot, training_model: TrainingModel, main_model: MainModel
) -> None:
    """
    Test to see if radio selection enables/disabled the model size correctly
    """
    # ARRANGE
    training_view: TrainingView = TrainingView(
        main_model=main_model,
        experiments_model=FakeExperimentsModel(),
        training_model=training_model,
        viewer=FakeViewer(),
    )
    # Fake a selection
    training_view._model_size_combo_box.setCurrentIndex(1)
    training_view._model_size_combo_box.setCurrentText("small")

    # Some checks before testing
    assert training_view._model_size_combo_box.isEnabled()
    assert training_model.get_model_size() == ModelSize.SMALL

    # ACT (click yes radio and disable model size selection)
    with qtbot.waitSignal(training_view.existing_model_yes_radio.clicked):
        training_view.existing_model_yes_radio.click()

    # ASSERT
    # model size selection now disabled and model is reset
    assert not training_view._model_size_combo_box.isEnabled()
    assert training_view._model_size_combo_box.currentIndex() == -1
    assert training_model.get_model_size() is None

    # ACT (click no radio again and re-enable model size selection)
    with qtbot.waitSignal(training_view.existing_model_no_radio.clicked):
        training_view.existing_model_no_radio.click()

    # ASSERT
    assert training_view._model_size_combo_box.isEnabled()



# from pathlib import Path
from pathlib import Path
import pytest
from allencell_ml_segmenter._tests.fakes.fake_subscriber import FakeSubscriber
from allencell_ml_segmenter._tests.fakes.fake_user_settings import (
    FakeUserSettings,
)
from allencell_ml_segmenter.config.i_user_settings import IUserSettings

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
import allencell_ml_segmenter


@pytest.fixture
def experiments_model() -> ExperimentsModel:
    exp_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "main"
        / "experiments_home"
    )
    experiments_model = ExperimentsModel(
        FakeUserSettings(
            cyto_dl_home_path=Path(), user_experiments_path=exp_path
        )
    )
    experiments_model.apply_experiment_name("2_exp")
    return experiments_model


def test_refresh_experiments(experiments_model: ExperimentsModel) -> None:
    expected_sorted = [
        "1_exp",
        "2_exp",
        "a_exp",
        "one_ckpt_exp",
        "z_exp",
    ]  # Only these directories have checkpoint sub dirs.
    experiments_model.refresh_experiments()
    assert len(experiments_model.get_experiments()) == len(expected_sorted)
    assert experiments_model.get_experiments() == expected_sorted


def test_get_cyto_dl_config() -> None:
    expected_config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=Path(__file__).parent / "experiments_home",
    )
    model = ExperimentsModel(expected_config)
    assert model.get_user_settings() == expected_config


def test_get_user_experiments_path() -> None:
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    assert model.get_user_experiments_path() == user_experiments_path


def test_get_model_checkpoints() -> None:
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    expected = user_experiments_path / "foo" / "checkpoints" / "bar"
    model = ExperimentsModel(config)
    assert model.get_model_checkpoints_path("foo", "bar") == expected


def test_get_model_checkpoints_no_experiment_name() -> None:
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)

    with pytest.raises(ValueError):
        model.get_model_checkpoints_path(None, "bar")


def test_get_model_checkpoints_no_checkpoint() -> None:
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)

    with pytest.raises(ValueError):
        model.get_model_checkpoints_path("foo", None)


def test_get_train_config_path() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    expected = user_experiments_path / "test_experiment" / "train_config.yaml"
    model = ExperimentsModel(config)
    model.apply_experiment_name("test_experiment")

    # Act / Assert
    assert model.get_train_config_path() == expected


def test_get_csv_path() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    model.apply_experiment_name("0_exp")
    expected = user_experiments_path / "0_exp" / "data"

    # Act / Assert
    assert model.get_csv_path() == expected


def test_get_metrics_csv_path() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    model.apply_experiment_name("0_exp")
    expected = user_experiments_path / "0_exp" / "csv"

    # Act / Assert
    assert model.get_metrics_csv_path() == expected


def test_get_latest_metrics_csv_version_no_versions() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    model.apply_experiment_name("0_exp")

    # Act / Assert
    assert model.get_latest_metrics_csv_version() == -1


def test_get_latest_metrics_csv_version_no_directory() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    model.apply_experiment_name("2_exp")

    # Act / Assert
    assert model.get_latest_metrics_csv_version() == -1


def test_get_latest_metrics_csv_version_version_1() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    model.apply_experiment_name("1_exp")

    # Act / Assert
    assert model.get_latest_metrics_csv_version() == 1


def test_get_current_epoch_with_ckpt() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    model.apply_experiment_name("one_ckpt_exp")

    # Act / Assert
    assert model.get_current_epoch() == 0


def test_get_current_epoch_no_ckpt() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    # this experiment has no checkpoints, so we expect current epoch to be undefined
    model.apply_experiment_name("0_exp")

    # Act / Assert
    assert model.get_current_epoch() is None


def test_apply_experiment_name() -> None:
    # Arrange
    user_experiments_path = Path(__file__).parent / "experiments_home"
    config = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model = ExperimentsModel(config)
    subscriber = FakeSubscriber()
    model.subscribe(
        Event.ACTION_EXPERIMENT_APPLIED, subscriber, subscriber.handle
    )
    expected = "test_experiment"

    # Act
    model.apply_experiment_name(expected)

    # Assert
    assert model.get_experiment_name() == expected
    assert subscriber.was_handled(Event.ACTION_EXPERIMENT_APPLIED)


def test_get_best_ckpt_experiment_selected() -> None:
    # ARRANGE
    user_experiments_path: Path = Path(__file__).parent / "experiments_home"
    config: FakeUserSettings = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model: ExperimentsModel = ExperimentsModel(config)
    model.apply_experiment_name("2_exp")

    # ACT
    selected_ckpt: str = model.get_best_ckpt()

    assert selected_ckpt.name == "1.ckpt"


def test_get_best_ckpt_from_current_experiment_no_experiment_selected():
    # ARRANGE
    user_experiments_path: Path = Path(__file__).parent / "experiments_home"
    config: FakeUserSettings = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model: ExperimentsModel = ExperimentsModel(config)

    # ACT
    selected_ckpt: str = (
        model.get_best_ckpt()
    )  # dont specify a specific experiemnt to grab the ckpt from

    assert selected_ckpt is None


def test_get_best_ckpt_no_checkpoint_dir():
    # ARRANGE
    user_experiments_path: Path = Path(__file__).parent / "experiments_home"
    config: FakeUserSettings = FakeUserSettings(
        cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
        user_experiments_path=user_experiments_path,
    )
    model: ExperimentsModel = ExperimentsModel(config)
    model.apply_experiment_name(
        "0_exp"
    )  # No checkpoint folder in this test dir

    # ACT
    selected_ckpt: str = model.get_best_ckpt()

    # assert
    assert selected_ckpt is None



import multiprocessing
import torch
import platform


class CUDAUtils:

    @staticmethod
    def cuda_available() -> bool:
        """
        Determines if the user has the correct CUDA drivers installed in order
        accelerate torch using a GPU
        """
        return torch.cuda.is_available()

    @staticmethod
    def get_num_workers() -> int:
        """
        Get the number of available cpu cores on this machine
        """
        # For MACOS or CPU runs:
        # On MACOS we cannot set num_workers no matter what.
        # On CPU, increasing num_workers will offer no performance increase
        #   as dataloading is not the bottleneck
        if platform.system() == "Darwin" or not CUDAUtils.cuda_available():
            return 0
        # For Windows/Linux:
        # We set num_workers to 1 for GPU runs
        # num_workers=1 should be able to support most systems while providing a small speed benefit.
        else:
            return 1



import pytest
import numpy as np

from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.thresholding.thresholding_model import (
    ThresholdingModel,
)
from allencell_ml_segmenter.thresholding.thresholding_service import (
    ThresholdingService,
)
from allencell_ml_segmenter.core.task_executor import SynchroTaskExecutor
from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer


@pytest.fixture
def test_image():
    """Create a small test image for thresholding."""
    return np.array([[10, 20, 30], [40, 50, 60], [70, 80, 90]])


def test_on_threshold_changed_non_prediction(test_image):
    # ARRANGE
    thresholding_model: ThresholdingModel = ThresholdingModel()
    viewer: FakeViewer = FakeViewer()
    thresholding_service: ThresholdingService = ThresholdingService(
        thresholding_model,
        FakeExperimentsModel(),
        MainModel(),
        viewer,
        task_executor=SynchroTaskExecutor.global_instance(),
    )
    viewer.add_image(test_image, name="test_layer")

    # ACT set a threshold to trigger
    thresholding_model.set_thresholding_value(50)

    # Verify a segmentation layer is added
    assert "[threshold] test_layer" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] test_layer"]
    assert np.array_equal(seg_data, (test_image > 50).astype(int))

    # check if existing thresholds get updated
    thresholding_model.set_thresholding_value(100)
    assert len(viewer.get_layers()) == 1
    seg_data = viewer.threshold_inserted["[threshold] test_layer"]
    assert np.array_equal(seg_data, (test_image > 100).astype(int))


def test_on_threshold_changed_non_prediction(test_image):
    """
    Test that the thresholding service does not add a threshold layer for a layer that is not a probability map
    """
    # ARRANGE
    thresholding_model: ThresholdingModel = ThresholdingModel()
    viewer: FakeViewer = FakeViewer()
    main_model: MainModel = MainModel()
    main_model.set_predictions_in_viewer(True)
    thresholding_service: ThresholdingService = ThresholdingService(
        thresholding_model,
        FakeExperimentsModel(),
        main_model,
        viewer,
        task_executor=SynchroTaskExecutor.global_instance(),
    )

    # Only the [seg] layers below should produce a threshold layer since they have prob map metadata
    viewer.add_image(test_image, name="[raw] test_layer 1")
    viewer.add_image(
        test_image,
        name="[seg] test_layer 1",
        metadata={"prob_map": test_image},
    )
    viewer.add_image(test_image, name="[raw] test_layer 2")
    viewer.add_image(
        test_image,
        name="[seg] test_layer 2",
        metadata={"prob_map": test_image},
    )
    viewer.add_image(test_image, name="donotthreshold")
    thresholding_model.set_thresholding_layers(
        viewer.get_all_layers_containing_prob_map()
    )
    thresholding_model.set_selected_idx([0, 1])

    # ACT set a threshold to trigger
    thresholding_model.set_thresholding_value(50)
    thresholding_model.set_threshold_enabled(True)

    # Verify a threshold layer is added for each seg layer
    assert "[threshold] [seg] test_layer 1" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] [seg] test_layer 1"]
    assert np.array_equal(seg_data, (test_image > 50).astype(int))
    assert "[threshold] [seg] test_layer 2" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] [seg] test_layer 2"]
    assert np.array_equal(seg_data, (test_image > 50).astype(int))
    # verify that raw layers do not get thresholded
    assert len(viewer.threshold_inserted) == 2

    # verify existing threshold layers get updated correctly
    thresholding_model.set_thresholding_value(100)
    # Verify a threshold layer is added for each seg layer
    assert "[threshold] [seg] test_layer 1" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] [seg] test_layer 1"]
    assert np.array_equal(seg_data, (test_image > 100).astype(int))
    assert "[threshold] [seg] test_layer 2" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] [seg] test_layer 2"]
    assert np.array_equal(seg_data, (test_image > 100).astype(int))
    # verify that raw layers do not get thresholded
    assert len(viewer.threshold_inserted) == 2


def test_none_selected_removes_binary_maps(test_image):
    """
    Test that the thresholding service does not add a threshold layer for a layer that is not a probability map
    """
    # ARRANGE
    thresholding_model: ThresholdingModel = ThresholdingModel()
    viewer: FakeViewer = FakeViewer()
    main_model: MainModel = MainModel()
    main_model.set_predictions_in_viewer(True)
    thresholding_service: ThresholdingService = ThresholdingService(
        thresholding_model,
        FakeExperimentsModel(),
        main_model,
        viewer,
        task_executor=SynchroTaskExecutor.global_instance(),
    )

    # Only the [seg] layers below should produce a threshold layer since they have prob map metadata
    viewer.add_image(test_image, name="[raw] test_layer 1")
    viewer.add_image(
        test_image,
        name="[seg] test_layer 1",
        metadata={"prob_map": test_image},
    )
    viewer.add_image(test_image, name="[raw] test_layer 2")
    viewer.add_image(
        test_image,
        name="[seg] test_layer 2",
        metadata={"prob_map": test_image},
    )
    viewer.add_image(test_image, name="donotthreshold")
    thresholding_model.set_thresholding_layers(
        viewer.get_all_layers_containing_prob_map()
    )
    thresholding_model.set_selected_idx([0, 1])

    # create binary maps and add to viewer
    thresholding_model.set_thresholding_value(50)
    thresholding_model.set_threshold_enabled(True)

    # Verify a threshold layer is added for each seg layer
    assert "[threshold] [seg] test_layer 1" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] [seg] test_layer 1"]
    assert np.array_equal(seg_data, (test_image > 50).astype(int))
    assert "[threshold] [seg] test_layer 2" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] [seg] test_layer 2"]
    assert np.array_equal(seg_data, (test_image > 50).astype(int))
    # verify that raw layers do not get thresholded
    assert len(viewer.threshold_inserted) == 2

    # ACT remove all binary maps by selecting None
    thresholding_model.disable_all()

    # Assert
    assert len(viewer.threshold_inserted) == 0


def test_unselection_removes_binary_maps(test_image):
    """
    Test that the thresholding service does not add a threshold layer for a layer that is not a probability map
    """
    # ARRANGE
    thresholding_model: ThresholdingModel = ThresholdingModel()
    viewer: FakeViewer = FakeViewer()
    main_model: MainModel = MainModel()
    main_model.set_predictions_in_viewer(True)
    thresholding_service: ThresholdingService = ThresholdingService(
        thresholding_model,
        FakeExperimentsModel(),
        main_model,
        viewer,
        task_executor=SynchroTaskExecutor.global_instance(),
    )

    # Only the [seg] layers below should produce a threshold layer since they have prob map metadata
    viewer.add_image(test_image, name="[raw] test_layer 1")
    viewer.add_image(
        test_image,
        name="[seg] test_layer 1",
        metadata={"prob_map": test_image},
    )
    viewer.add_image(test_image, name="[raw] test_layer 2")
    viewer.add_image(
        test_image,
        name="[seg] test_layer 2",
        metadata={"prob_map": test_image},
    )
    viewer.add_image(test_image, name="donotthreshold")
    thresholding_model.set_thresholding_layers(
        viewer.get_all_layers_containing_prob_map()
    )
    thresholding_model.set_selected_idx([0, 1])

    # create binary maps and add to viewer
    thresholding_model.set_thresholding_value(50)
    thresholding_model.set_threshold_enabled(True)

    # Verify a threshold layer is added for each seg layer
    assert "[threshold] [seg] test_layer 1" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] [seg] test_layer 1"]
    assert np.array_equal(seg_data, (test_image > 50).astype(int))
    assert "[threshold] [seg] test_layer 2" in viewer.threshold_inserted
    seg_data = viewer.threshold_inserted["[threshold] [seg] test_layer 2"]
    assert np.array_equal(seg_data, (test_image > 50).astype(int))
    # verify that raw layers do not get thresholded
    assert len(viewer.threshold_inserted) == 2

    # ACT remove binary map by unselecting index 1
    thresholding_model.set_selected_idx(
        [0]
    )  # remove 1, which was previously selected

    # Assert
    assert len(viewer.threshold_inserted) == 1
    assert "[threshold] [seg] test_layer 2" not in viewer.threshold_inserted
    assert "[threshold] [seg] test_layer 1" in viewer.threshold_inserted



from pathlib import Path

from allencell_ml_segmenter.utils.zip_file import IZipFileManager


class FakeZipFileManager(IZipFileManager):
    """
    FakeZipFileManager for testing
    """

    def __init__(self) -> None:
        super().__init__()
        # zip files that have been written, keys- path, values- content written
        self.written_zip_files: dict[Path, bytes] = {}
        # list of unzipped files
        self.unzipped_files: list[Path] = []

    def write_zip_file(self, path: Path, contents: bytes) -> None:
        self.written_zip_files[path] = contents

    def unzip_zipped_file_and_delete_zip(self, path_to_zipped: Path) -> None:
        self.unzipped_files.append(path_to_zipped)



from pathlib import Path
import requests  # type: ignore
from typing import Optional

from allencell_ml_segmenter.core.dialog_box import DialogBox
from allencell_ml_segmenter.utils.s3.s3_request_exception import (
    S3RequestException,
)
from allencell_ml_segmenter.utils.zip_file import (
    IZipFileManager,
    ZipFileManager,
)


class AvailableModel:
    def __init__(
        self,
        model_file_name: str,
        bucket_endpoint: str,
        path_to_save_model: Path,
        zip_file_manager: IZipFileManager = ZipFileManager(),
    ):
        self._name: str = model_file_name
        # the url of this AvailableModel on the specified s3 bucket
        self._object_url: str = f"{bucket_endpoint}/{model_file_name}"
        # ZipFileManager handles all zip-file related operations for the model
        self._zipfile: IZipFileManager = zip_file_manager
        # dir where models should be downloaded to
        self._path_to_store_model: Path = path_to_save_model

    def download_model_and_unzip(self) -> None:
        """
        Download this AvilableModel from s3 and unzip it into the specified :param path:
        """
        continue_download: bool = True
        # check to see if model already exists.
        if (self._path_to_store_model / Path(self.get_name()).stem).exists():
            overwrite_dialog = DialogBox(
                f"{self.get_name()} is already in your experiments folder. Overwrite?"
            )
            overwrite_dialog.exec()
            selection: Optional[bool] = overwrite_dialog.get_selection()
            continue_download = selection if selection is not None else False

        if continue_download:
            response: requests.Response = requests.get(self._object_url)

            if response.status_code == 200:
                # Write contents of s3:getObject to the path (writing the zip file)
                self._zipfile.write_zip_file(
                    self._path_to_store_model / self._name, response.content
                )
                # Unzip the contents of the zipfile where it is located, and delete the .zip
                self._zipfile.unzip_zipped_file_and_delete_zip(
                    self._path_to_store_model / self._name
                )
            else:
                # Something went wrong when downloading the object from s3
                raise S3RequestException(
                    f"Could not download model named {self._name} from {self._object_url}. Failed with status code {response.status_code}"
                )

    def get_name(self) -> str:
        return self._name

    def get_object_url(self) -> str:
        return self._object_url



import numpy as np
from pathlib import Path
from typing import List, Optional, Dict, Any
from enum import Enum

from qtpy.QtCore import Signal, QObject

from allencell_ml_segmenter.curation.curation_data_class import CurationRecord
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel, ImageType
from allencell_ml_segmenter.core.image_data_extractor import ImageData


class CurationView(Enum):
    INPUT_VIEW = "input_view"
    MAIN_VIEW = "main_view"


class CurationModel(QObject):
    """
    Stores state relevant to prediction processes.
    """

    current_view_changed: Signal = Signal()
    image_directory_set: Signal = Signal(ImageType)
    channel_count_set: Signal = Signal(ImageType)

    cursor_moved: Signal = Signal()
    image_loading_finished: Signal = Signal()

    save_to_disk_requested: Signal = Signal()
    saved_to_disk: Signal = Signal(bool)

    def __init__(
        self,
        experiments_model: ExperimentsModel,
        main_model: MainModel,
    ) -> None:
        super().__init__()
        # should always start at input view
        self._experiments_model: ExperimentsModel = experiments_model
        self._main_model: MainModel = main_model
        self._current_view: CurationView = CurationView.INPUT_VIEW

        self._img_dirs: Dict[ImageType, Optional[Path]] = (
            self._get_placeholder_dict()
        )
        self._img_dir_paths: Dict[ImageType, Optional[List[Path]]] = (
            self._get_placeholder_dict()
        )

        # These are what the user has selected in the input view
        self._selected_channels: Dict[ImageType, Optional[int]] = (
            self._get_placeholder_dict()
        )
        # these are the total number of channels for the images in the folder
        self._channel_counts: Dict[ImageType, Optional[int]] = (
            self._get_placeholder_dict()
        )

        self._curation_record: Optional[List[CurationRecord]] = None
        # None until start_image_loading is called
        self._cursor: Optional[int] = None
        # True when images have been dropped from memory
        self._image_loading_stopped: bool = False
        # private invariant: _next_img_data will only have < self._get_num_data_dict_keys() keys if
        # a thread is currently updating _next_img_data. Same goes for curr
        self._curr_img_data: Optional[dict[ImageType, Optional[ImageData]]] = (
            None
        )
        self._next_img_data: Optional[dict[ImageType, Optional[ImageData]]] = (
            None
        )

    def get_merging_mask(self) -> Optional[np.ndarray]:
        return (
            self._curation_record[self._cursor].merging_mask
            if self._curation_record is not None and self._cursor is not None
            else None
        )

    def set_merging_mask(self, mask: Optional[np.ndarray]) -> None:
        if self._curation_record is not None and self._cursor is not None:
            self._curation_record[self._cursor].merging_mask = mask

    def get_excluding_mask(self) -> Optional[np.ndarray]:
        return (
            self._curation_record[self._cursor].excluding_mask
            if self._curation_record is not None and self._cursor is not None
            else None
        )

    def set_excluding_mask(self, mask: Optional[np.ndarray]) -> None:
        if self._curation_record is not None and self._cursor is not None:
            self._curation_record[self._cursor].excluding_mask = mask

    def get_base_image(self) -> Optional[str]:
        return (
            self._curation_record[self._cursor].base_image
            if self._curation_record is not None and self._cursor is not None
            else None
        )

    def set_base_image(self, base: Optional[str]) -> None:
        if self._curation_record is not None and self._cursor is not None:
            self._curation_record[self._cursor].base_image = base

    def get_use_image(self) -> Optional[bool]:
        return (
            self._curation_record[self._cursor].to_use
            if self._curation_record is not None and self._cursor is not None
            else None
        )

    def set_use_image(self, use: bool) -> None:
        if self._curation_record is not None and self._cursor is not None:
            self._curation_record[self._cursor].to_use = use

    def set_image_directory(self, img_type: ImageType, dir: Path) -> None:
        self._img_dirs[img_type] = dir
        self.image_directory_set.emit(img_type)

    def get_image_directory(self, img_type: ImageType) -> Optional[Path]:
        return self._img_dirs[img_type]

    def set_image_directory_paths(
        self, img_type: ImageType, paths: List[Path]
    ) -> None:
        self._img_dir_paths[img_type] = paths

    def get_image_directory_paths(
        self, img_type: ImageType
    ) -> Optional[List[Path]]:
        return self._img_dir_paths[img_type]

    def set_selected_channel(self, img_type: ImageType, channel: int) -> None:
        self._selected_channels[img_type] = channel

    def get_selected_channel(self, img_type: ImageType) -> Optional[int]:
        return self._selected_channels[img_type]

    def set_current_view(self, view: CurationView) -> None:
        """
        Set current curation view
        """
        # TODO: reset all state? only relevant if we expect a nonlinear path through curation, or multiple curation
        # rounds in a single session
        if view != self._current_view:
            if view == CurationView.MAIN_VIEW:
                self._curation_record = self._generate_new_curation_record()
                seg2_exists: bool = (
                    self._img_dir_paths[ImageType.SEG2] is not None
                )
                self._curr_img_data = self._get_placeholder_dict(
                    incl_seg2=seg2_exists
                )
                self._next_img_data = self._get_placeholder_dict(
                    incl_seg2=seg2_exists
                )
                self._curation_record_saved_to_disk = False
                # set the central selected channels for the app once the user clicks 'start curation'
                self._main_model.set_selected_channels(self._selected_channels)
            else:
                self._curation_record = None
                self._curr_img_data = None
                self._next_img_data = None
            self._current_view = view
            self.current_view_changed.emit()

    def get_current_view(self) -> CurationView:
        """
        Get current curation view
        """
        return self._current_view

    def set_channel_count(self, img_type: ImageType, count: int) -> None:
        self._channel_counts[img_type] = count
        self.channel_count_set.emit(img_type)

    def get_channel_count(self, img_type: ImageType) -> Optional[int]:
        return self._channel_counts[img_type]

    def get_save_masks_path(self) -> Optional[Path]:
        exp_name: Optional[str] = self._experiments_model.get_experiment_name()
        exp_path: Optional[Path] = (
            self._experiments_model.get_user_experiments_path()
        )
        return (
            exp_path / exp_name
            if exp_name is not None and exp_path is not None
            else None
        )

    def get_curation_record(self) -> Optional[list[CurationRecord]]:
        return self._curation_record

    # WARNING: methods that access data dicts must only be called from the main thread
    def set_curr_image_data(
        self, img_type: ImageType, img_data: ImageData
    ) -> None:
        if self._curr_img_data is None:
            raise RuntimeError("Current image data is uninitialized")
        self._curr_img_data[img_type] = img_data
        if not self.is_waiting_for_images():
            self.image_loading_finished.emit()

    def get_curr_image_data(self, img_type: ImageType) -> Optional[ImageData]:
        return (
            self._curr_img_data[img_type]
            if self._curr_img_data is not None
            else None
        )

    def set_next_image_data(
        self, img_type: ImageType, img_data: ImageData
    ) -> None:
        if self._next_img_data is None:
            raise RuntimeError("Next image data is uninitialized")
        self._next_img_data[img_type] = img_data
        if not self.is_waiting_for_images():
            self.image_loading_finished.emit()

    # note: I don't see a reason why we would need to get the next image data instead of
    # calling next_image, so leaving that out
    def has_seg2_data(self) -> bool:
        return (
            self._img_dir_paths[ImageType.SEG2] is not None
            if self._img_dir_paths is not None
            else False
        )

    def get_num_images(self) -> int:
        if (
            self._img_dir_paths is not None
            and self._img_dir_paths[ImageType.RAW] is not None
        ):
            return len(self._img_dir_paths[ImageType.RAW])  # type: ignore
        return 0

    def get_curr_image_index(self) -> Optional[int]:
        return self._cursor

    def has_next_image(self) -> bool:
        if self._cursor is None:
            return False
        return self._cursor + 1 < self.get_num_images()

    def is_waiting_for_curr_images(self) -> bool:
        if self._curr_img_data is None:
            return False
        return len(self._curr_img_data) != self._get_num_data_dict_keys()

    def is_waiting_for_next_images(self) -> bool:
        if self._next_img_data is None:
            return False
        return len(self._next_img_data) != self._get_num_data_dict_keys()

    def is_waiting_for_images(self) -> bool:
        return (
            self.is_waiting_for_curr_images()
            or self.is_waiting_for_next_images()
        )

    def get_image_loading_stopped(self) -> bool:
        return self._image_loading_stopped

    def start_loading_images(self) -> None:
        """
        Must be called before attempting to get image data.
        Signals emitted:
        immediate: cursor_moved
        at some point: image_loading_finished
        """
        if self._curr_img_data is None or self._next_img_data is None:
            raise RuntimeError(
                "Cannot start loading when image data dict is uninitialized"
            )

        self._cursor = 0
        # need to set use image to true since we want this to be the default
        self.set_use_image(True)
        self._curr_img_data.clear()
        if self.has_next_image():
            self._next_img_data.clear()
        self.cursor_moved.emit()

    def stop_loading_images(self) -> None:
        """
        Drops pre-loaded images from memory and prevents further
        image loading from occurring in curation.
        """
        if self._curr_img_data is None or self._next_img_data is None:
            raise RuntimeError(
                "Cannot stop loading when image data dict is uninitialized"
            )

        self._image_loading_stopped = True
        self._curr_img_data.clear()
        self._next_img_data.clear()

    def next_image(self) -> None:
        """
        Move to the next image for curation.
        Signals emitted:
        immediate: cursor_moved
        at some point: image_loading_finished
        """
        if self.get_image_loading_stopped():
            raise RuntimeError("Image loader is stopped.")
        if self.is_waiting_for_images():
            raise RuntimeError(
                "Image loader is busy. Please see image_loading_finished signal."
            )
        if not self.has_next_image() or self._cursor is None:
            raise RuntimeError("No next image available")

        self._curr_img_data = self._next_img_data
        self._cursor += 1
        self._next_img_data = (
            {}
            if self.has_next_image()
            else self._get_placeholder_dict(incl_seg2=self.has_seg2_data())
        )
        # need to set use image to true since we want this to be the default
        self.set_use_image(True)
        self.cursor_moved.emit()
        # client expects that calling next will eventually result in a image_loading_finished signal
        if not self.has_next_image():
            self.image_loading_finished.emit()

    def set_curation_record_saved_to_disk(self, saved: bool) -> None:
        self.saved_to_disk.emit(saved)

    def save_curr_curation_record_to_disk(self) -> None:
        self.save_to_disk_requested.emit()

    def get_csv_path(self) -> Optional[Path]:
        exp_name: Optional[str] = self._experiments_model.get_experiment_name()
        exp_path: Optional[Path] = (
            self._experiments_model.get_user_experiments_path()
        )
        return (
            exp_path / exp_name / "data"
            if exp_name is not None and exp_path is not None
            else None
        )

    def get_max_num_images_to_use(self) -> int:
        """
        Returns the maximum size of the resulting dataset at this point in curation. For example,
        if the user has 12 images and has marked 3 as 'do not use', this method will return 9.
        Note that the current 'to_use' selection is not included, as it is not yet finalized.
        """
        if self._cursor is None or self._curation_record is None:
            raise RuntimeError(
                "Cannot calculate with undefined cursor or curation record"
            )

        possible: int = self.get_num_images()
        for i in range(self._cursor):
            if not self._curation_record[i].to_use:
                possible -= 1
        return possible

    def get_num_images_selected_to_use(self) -> int:
        """
        Returns the number of images that the user has marked as 'use this image'. Current
        'to_use' selection is included.
        """
        return (
            sum([1 if rec.to_use else 0 for rec in self._curation_record])
            if self._curation_record is not None
            else 0
        )

    def _generate_new_curation_record(self) -> List[CurationRecord]:
        """
        Returns a list of curation records populated with the file paths in self._img_dir_paths. See
        CurationRecord constructor below for the default values.
        """
        raw_paths: Optional[list[Path]] = self._img_dir_paths[ImageType.RAW]
        seg1_paths: Optional[list[Path]] = self._img_dir_paths[ImageType.SEG1]
        if raw_paths is None or seg1_paths is None:
            raise RuntimeError("Expected values for raw and seg1 images")

        seg2_paths: Optional[list[Path]] = self._img_dir_paths[ImageType.SEG2]
        if len(raw_paths) != len(seg1_paths) or (
            seg2_paths is not None and len(seg1_paths) != len(seg2_paths)
        ):
            raise ValueError("provided image dirs must be of same length")
        elif len(raw_paths) < 1:
            raise ValueError("cannot load images from empty image dir")

        return [
            CurationRecord(
                raw_paths[i],
                seg1_paths[i],
                (seg2_paths[i] if seg2_paths is not None else None),
                None,
                None,
                "seg1",
                False,
            )
            for i in range(len(raw_paths))
        ]

    def _get_num_data_dict_keys(self) -> int:
        """
        Returns expected number of keys in an img data dict that is not being written to
        asynchronously.
        """
        return 3 if self.has_seg2_data() else 2

    def _get_placeholder_dict(
        self, incl_seg2: bool = True
    ) -> Dict[ImageType, Optional[Any]]:
        """
        Returns placeholder image data dict with keys mapped to None. Only includes a SEG2 key if
        :param incl_seg2: is True.
        """
        output: Dict[ImageType, Optional[Any]] = {
            ImageType.RAW: None,
            ImageType.SEG1: None,
        }
        if incl_seg2:
            output[ImageType.SEG2] = None
        return output



from pathlib import Path

import pytest
import responses

import allencell_ml_segmenter
from allencell_ml_segmenter.utils.s3.s3_available_model import AvailableModel
from allencell_ml_segmenter.utils.s3.s3_model_bucket import (
    S3ModelBucket,
)
from allencell_ml_segmenter.utils.s3.s3_request_exception import (
    S3RequestException,
)
from allencell_ml_segmenter._tests.utils.s3.s3_response_fixtures import (
    s3_response_listobjectv2_contents_two_models,
    s3_response_listobjectv2_contents_duplicate_model,
)
from allencell_ml_segmenter.utils.s3.s3_bucket_constants import PROD_BUCKET


@responses.activate
def test_get_available_models(
    s3_response_listobjectv2_contents_two_models: str,
) -> None:
    # add fake xml response that is returned when we make a request to the test_url with the list-type=2 param
    responses.add(
        **{
            "method": responses.GET,
            "url": f"{PROD_BUCKET}?list-type=2",
            "body": s3_response_listobjectv2_contents_two_models,
            "status": 200,
            "content_type": "application/xml",
            "adding_headers": {"X-Foo": "Bar"},
        }
    )
    # Test path to save zip to
    test_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "zip_files"
    )

    model_downloader: S3ModelBucket = S3ModelBucket(PROD_BUCKET, test_path)

    # ACT
    available_models_dict: dict[str, AvailableModel] = (
        model_downloader.get_available_models()
    )

    # ASSERT
    # should have 2 models available
    assert len(available_models_dict.keys()) == 2
    assert "model1" in available_models_dict.keys()
    assert "model2" in available_models_dict.keys()
    assert "some_random_file" not in available_models_dict.keys()

    # dictionary values contain available models object w/ name and object url on s3
    # check if object url was built correctly for the AvailableModels
    assert available_models_dict["model1"].get_name() == "model1.zip"
    assert (
        available_models_dict["model1"].get_object_url()
        == f"{PROD_BUCKET}/model1.zip"
    )
    assert available_models_dict["model2"].get_name() == "model2.zip"
    assert (
        available_models_dict["model2"].get_object_url()
        == f"{PROD_BUCKET}/model2.zip"
    )


@responses.activate
def test_get_available_models_duplicate_file_error(
    s3_response_listobjectv2_contents_duplicate_model: str,
) -> None:
    # ARRANGE
    # add fake xml response that is returned when we make a request to the test_url with the list-type=2 param
    responses.add(
        **{
            "method": responses.GET,
            "url": f"{PROD_BUCKET}?list-type=2",
            "body": s3_response_listobjectv2_contents_duplicate_model,
            "status": 200,
            "content_type": "application/xml",
            "adding_headers": {"X-Foo": "Bar"},
        }
    )
    # Test path to save zip to
    test_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "zip_files"
    )

    model_downloader: S3ModelBucket = S3ModelBucket(PROD_BUCKET, test_path)

    # ACT/ASSERT
    with pytest.raises(ValueError):
        # if there are models with duplicate names on s3 somehow- we show an error
        model_downloader.get_available_models()


@responses.activate
def test_get_available_models_bad_request() -> None:
    # ARRANGE
    error_response: str = "error"
    responses.add(
        **{
            "method": responses.GET,
            "url": f"{PROD_BUCKET}?list-type=2",
            "body": error_response,
            "status": 400,
            "content_type": "error",
            "adding_headers": {"X-Foo": "Bar"},
        }
    )
    # Test path to save zip to
    test_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "zip_files"
    )
    model_downloader: S3ModelBucket = S3ModelBucket(PROD_BUCKET, test_path)

    # ACT/ASSERT
    with pytest.raises(S3RequestException):
        model_downloader.get_available_models()  # any status code != 200 should throw S3RequestException



from qtpy.QtCore import QSettings


class FakeQSettings(QSettings):
    def __init__(self):
        self.keys = {}

    def value(self, key):
        return self.keys[key]

    def setValue(self, key, value):
        self.keys[key] = value



from pathlib import Path
import pytest

import allencell_ml_segmenter
from allencell_ml_segmenter.utils.file_utils import FileUtils
from allencell_ml_segmenter.utils.file_writer import (
    IFileWriter,
    FakeFileWriter,
)
from allencell_ml_segmenter.curation.curation_data_class import CurationRecord

from typing import List, Set
import numpy as np


def test_get_all_files_in_dir() -> None:
    # arrange
    folder: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "img_folder"
    )

    # act
    all_files = FileUtils.get_all_files_in_dir_ignore_hidden(folder)

    # assert
    assert len(all_files) == 5
    assert all_files[0].name == "t1.tiff"
    assert all_files[1].name == "t2.tiff"
    assert all_files[2].name == "t3.tiff"
    assert all_files[3].name == "t4.tiff"
    assert all_files[4].name == "t5.tiff"


def test_get_all_files_in_dir_with_hidden_files() -> None:
    # arrange
    folder: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "img_folder_with_hidden_files"
    )

    # act
    all_files = FileUtils.get_all_files_in_dir_ignore_hidden(folder)

    # assert
    assert len(all_files) == 1
    assert all_files[0].name == "t1.tiff"


def test_get_img_path_from_folder():
    folder: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "img_folder"
    )
    img: Path = FileUtils.get_img_path_from_folder(folder)
    assert any(
        [
            img.samefile(exist_img)
            for exist_img in [folder / f"t{i + 1}.tiff" for i in range(5)]
        ]
    )


def test_get_img_path_from_folder_hidden_files():
    folder: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "img_folder_with_hidden_files"
    )
    t1: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "img_folder_with_hidden_files"
        / "t1.tiff"
    )
    img: Path = FileUtils.get_img_path_from_folder(folder)
    assert img.samefile(t1)


FAKE_CSV_PATH: Path = Path("fakecsv").resolve()
FAKE_MASK_PATH: Path = Path("fakemask").resolve()
EXP_TRAIN_PATH: Path = FAKE_CSV_PATH / "train.csv"
EXP_TEST_PATH: Path = FAKE_CSV_PATH / "test.csv"
EXP_VAL_PATH: Path = FAKE_CSV_PATH / "val.csv"


def _generate_default_records(num_records: int) -> List[CurationRecord]:
    cr: List[CurationRecord] = []
    for i in range(num_records):
        cr.append(
            CurationRecord(
                Path(f"raw_{i}"),
                Path(f"seg1_{i}"),
                Path(f"seg2_{i}"),
                None,
                None,
                "seg1",
                True,
            )
        )
    return cr


def test_write_curation_record_closes_files():
    # Arrange
    fake_writer: IFileWriter = FakeFileWriter()
    f_utils: FileUtils = FileUtils(fake_writer)
    fake_curation_record: List[CurationRecord] = _generate_default_records(10)

    # Act
    f_utils.write_curation_record(
        fake_curation_record, FAKE_CSV_PATH, FAKE_MASK_PATH
    )

    # Assert
    assert not fake_writer.csv_state[EXP_TRAIN_PATH]["open"]
    assert not fake_writer.csv_state[EXP_TEST_PATH]["open"]
    assert not fake_writer.csv_state[EXP_VAL_PATH]["open"]


def test_write_curation_record_writes_mask_to_disk():
    """
    Verify that a call to the public write_curation_record will result in excluding/merging masks being written to disk
    """
    # Arrange
    fake_writer: IFileWriter = FakeFileWriter()
    f_utils: FileUtils = FileUtils(fake_writer)
    fake_curation_record: List[CurationRecord] = _generate_default_records(4)
    masks: List[np.ndarray] = [
        np.asarray([[[1, 2], [3, 4], [5, 6]]]),
        np.asarray([[[2, 2], [3, 4], [5, 6]]]),
        np.asarray([[[3, 2], [3, 4], [5, 6]]]),
        np.asarray([[[4, 2], [3, 4], [5, 6]]]),
    ]
    # first record has just an excluding mask, second just a merging mask, third both
    fake_curation_record[0].excluding_mask = masks[0]
    fake_curation_record[1].merging_mask = masks[1]
    fake_curation_record[2].excluding_mask = masks[2]
    fake_curation_record[2].merging_mask = masks[3]

    expected_save_paths: List[Path] = [
        FAKE_MASK_PATH / "excluding_masks" / "excluding_mask_raw_0.npy",
        FAKE_MASK_PATH / "merging_masks" / "merging_mask_raw_1.npy",
        FAKE_MASK_PATH / "excluding_masks" / "excluding_mask_raw_2.npy",
        FAKE_MASK_PATH / "merging_masks" / "merging_mask_raw_2.npy",
    ]

    # Act
    f_utils.write_curation_record(
        fake_curation_record, FAKE_CSV_PATH, FAKE_MASK_PATH
    )

    # Assert
    assert len(fake_writer.np_save_state) == len(expected_save_paths)
    # verify that the expected mask was saved to the expected save path
    for mask, save_path in zip(masks, expected_save_paths):
        assert save_path in fake_writer.np_save_state
        assert np.array_equal(mask, fake_writer.np_save_state[save_path])


def test_write_curation_record_split_sizes():
    """
    Test that the split sizes are as expected for curation records
    of different lengths.
    """
    # Arrange
    fake_writer: IFileWriter = FakeFileWriter()
    f_utils: FileUtils = FileUtils(fake_writer)

    # Act / Assert
    with pytest.raises(RuntimeError) as e:
        f_utils.write_curation_record(
            _generate_default_records(1), FAKE_CSV_PATH, FAKE_MASK_PATH
        )

    with pytest.raises(RuntimeError) as e:
        f_utils.write_curation_record(
            _generate_default_records(3), FAKE_CSV_PATH, FAKE_MASK_PATH
        )

    with pytest.raises(RuntimeError) as e:
        records: list[CurationRecord] = _generate_default_records(4)
        records[0].to_use = False
        f_utils.write_curation_record(records, FAKE_CSV_PATH, FAKE_MASK_PATH)

    # Act
    f_utils.write_curation_record(
        _generate_default_records(4), FAKE_CSV_PATH, FAKE_MASK_PATH
    )

    # Assert
    assert len(fake_writer.csv_state[EXP_TRAIN_PATH]["rows"]) == 3
    assert len(fake_writer.csv_state[EXP_TEST_PATH]["rows"]) == 3
    assert len(fake_writer.csv_state[EXP_VAL_PATH]["rows"]) == 3

    # Act
    f_utils.write_curation_record(
        _generate_default_records(100), FAKE_CSV_PATH, FAKE_MASK_PATH
    )

    # Assert
    assert len(fake_writer.csv_state[EXP_TRAIN_PATH]["rows"]) == 91
    assert len(fake_writer.csv_state[EXP_TEST_PATH]["rows"]) == 11
    assert len(fake_writer.csv_state[EXP_VAL_PATH]["rows"]) == 11

    # Act
    f_utils.write_curation_record(
        _generate_default_records(1300), FAKE_CSV_PATH, FAKE_MASK_PATH
    )

    # Assert
    assert len(fake_writer.csv_state[EXP_TRAIN_PATH]["rows"]) == 1201
    assert len(fake_writer.csv_state[EXP_TEST_PATH]["rows"]) == 101
    assert len(fake_writer.csv_state[EXP_VAL_PATH]["rows"]) == 101


def test_write_curation_record_includes_only_selected_images():
    """
    Test that images marked as not to_use are not included in the records
    to be written.
    """
    # Arrange
    fake_writer: IFileWriter = FakeFileWriter()
    f_utils: FileUtils = FileUtils(fake_writer)

    fake_records: List[CurationRecord] = _generate_default_records(40)

    # mark half of the records as not to use
    for i in range(0, len(fake_records), 2):
        fake_records[i].to_use = False

    # Act
    f_utils.write_curation_record(fake_records, FAKE_CSV_PATH, FAKE_MASK_PATH)

    # Assert
    # there are 20 records to use, so we expect 18 for training 2 shared between test/val (plus header rows)
    assert len(fake_writer.csv_state[EXP_TRAIN_PATH]["rows"]) == 19
    assert len(fake_writer.csv_state[EXP_TEST_PATH]["rows"]) == 3
    assert len(fake_writer.csv_state[EXP_VAL_PATH]["rows"]) == 3


def test_write_curation_record_does_not_include_duplicates():
    """
    Test that the records in train and val are mutually exclusive. And that there
    are no duplicates within train or val.
    """
    # Arrange
    fake_writer: IFileWriter = FakeFileWriter()
    f_utils: FileUtils = FileUtils(fake_writer)

    fake_curation_record: List[CurationRecord] = _generate_default_records(20)

    # Act
    f_utils.write_curation_record(
        fake_curation_record, FAKE_CSV_PATH, FAKE_MASK_PATH
    )

    # Assert
    # expect that the rows in the training and val csv are mutually exclusive
    row_minus_index = lambda row: ",".join(row[1:])
    unique_rows: Set[str] = set()
    for row in fake_writer.csv_state[EXP_TRAIN_PATH]["rows"][1:]:
        assert row_minus_index(row) not in unique_rows
        unique_rows.add(row_minus_index(row))
    for row in fake_writer.csv_state[EXP_VAL_PATH]["rows"][1:]:
        assert row_minus_index(row) not in unique_rows
        unique_rows.add(row_minus_index(row))


def test_write_curation_record_test_val_same():
    """
    Test that the records in val and test are the same.
    """
    # Arrange
    fake_writer: IFileWriter = FakeFileWriter()
    f_utils: FileUtils = FileUtils(fake_writer)

    fake_curation_record: List[CurationRecord] = _generate_default_records(20)

    # Act
    f_utils.write_curation_record(
        fake_curation_record, FAKE_CSV_PATH, FAKE_MASK_PATH
    )

    # Assert
    # expect that rows in val and test are the same
    row_minus_index = lambda row: ",".join(row[1:])
    unique_rows: Set[str] = set()
    for row in fake_writer.csv_state[EXP_TEST_PATH]["rows"][1:]:
        unique_rows.add(row_minus_index(row))
    for row in fake_writer.csv_state[EXP_VAL_PATH]["rows"][1:]:
        assert row_minus_index(row) in unique_rows


def test_count_images_in_csv_folder_one_csv() -> None:
    # Arrange
    folder: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "csv"
    )

    # Act / Assert
    # expect one since there is only one unique raw file in this directory
    assert FileUtils.count_images_in_csv_folder(folder) == 1


def test_count_images_in_csv_folder_multiple_csv() -> None:
    """
    A more realistic test of the count function on a folder with a train, test, and val csv
    """
    # Arrange
    folder: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "multiple_csv"
    )

    # Act / Assert
    # expect 6 since train has 3 unique, test has 3 unique, val is a copy of test
    assert FileUtils.count_images_in_csv_folder(folder) == 6


def test_get_min_loss_from_csv() -> None:
    """
    Ensure the function returns the expected minimum loss from a well-formatted
    metrics csv file.
    """
    # Arrange
    csv_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "csv"
        / "metrics.csv"
    )

    # Act / Assert
    assert FileUtils.get_min_loss_from_csv(csv_path) == 0.9382843971252441


def test_get_min_loss_from_csv_invalid_csv() -> None:
    """
    Ensure the function returns none when the CSV doesn't contain the expected
    loss column.
    """
    # Arrange
    csv_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "csv"
        / "test_csv.csv"
    )

    # Act / Assert
    assert FileUtils.get_min_loss_from_csv(csv_path) is None


def test_get_min_loss_from_csv_empty_loss_col() -> None:
    """
    Ensure the function returns none when the CSV's loss column is empty.
    """
    # Arrange
    csv_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "csv"
        / "metrics_empty_loss.csv"
    )

    # Act / Assert
    assert FileUtils.get_min_loss_from_csv(csv_path) is None



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, List
from enum import Enum
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.publisher import Publisher


class InputMode(Enum):
    FROM_PATH = "from_path"
    FROM_NAPARI_LAYERS = "from_napari_layers"


class WidgetMode(Enum):
    PREDICTION = "prediction"
    THRESHOLDING = "thresholding"


class FileInputModel(Publisher):
    """
    Model for FileInputWidget
    """

    def __init__(self) -> None:
        super().__init__()
        self._input_image_path: Optional[Path] = None
        self._image_input_channel_index: Optional[int] = None
        self._output_directory: Optional[Path] = None
        self._input_mode: Optional[InputMode] = None
        self._selected_paths: Optional[list[Path]] = None
        self._max_channels: Optional[int] = None
        self._selected_idx: Optional[list[int]] = None

    def get_output_seg_directory(self) -> Optional[Path]:
        """
        Gets path to where segmentation predictions are stored.
        """
        output_dir: Optional[Path] = self.get_output_directory()
        if output_dir is None:
            return None
        return output_dir / "seg"

    def set_input_image_path(
        self, path: Optional[Path], extract_channels: bool = False
    ) -> None:
        self._input_image_path = path
        if extract_channels and path is not None:
            self.dispatch(Event.ACTION_FILEINPUT_EXTRACT_CHANNELS)

    def get_input_image_path(self) -> Optional[Path]:
        return self._input_image_path

    def set_image_input_channel_index(self, idx: Optional[int]) -> None:
        self._image_input_channel_index = idx

    def get_image_input_channel_index(self) -> Optional[int]:
        return self._image_input_channel_index

    def set_output_directory(self, dir: Optional[Path]) -> None:
        self._output_directory = dir

    def get_output_directory(self) -> Optional[Path]:
        return self._output_directory

    def set_input_mode(self, mode: Optional[InputMode]) -> None:
        self._input_mode = mode

    def get_input_mode(self) -> Optional[InputMode]:
        return self._input_mode

    def set_selected_paths(
        self, paths: Optional[List[Path]], extract_channels: bool = False
    ) -> None:
        self._selected_paths = paths
        if extract_channels and paths is not None:
            self.dispatch(Event.ACTION_FILEINPUT_EXTRACT_CHANNELS)

    def get_max_channels(self) -> Optional[int]:
        return self._max_channels

    def set_max_channels(self, max: Optional[int]) -> None:
        self._max_channels = max
        if max is not None:
            self.dispatch(Event.ACTION_FILEINPUT_MAX_CHANNELS_SET)

    def get_selected_paths(self) -> Optional[list[Path]]:
        return self._selected_paths

    def get_input_files_as_list(self) -> List[Path]:
        input_image_path: Optional[Path] = self.get_input_image_path()
        if (
            self.get_input_mode() == InputMode.FROM_PATH
            and input_image_path is not None
        ):
            return list(input_image_path.glob("*"))
        elif self.get_input_mode() == InputMode.FROM_NAPARI_LAYERS:
            selected_paths: Optional[list[Path]] = self.get_selected_paths()
            if selected_paths is not None:
                return selected_paths
        return []
